"""Utility functions for managing object state."""

import sys
import types

import six
from tippo import Any, Callable, Mapping, Type

from .dynamic_code import generate_unique_filename, make_function
from .get_mro import get_mro
from .import_path import get_path, import_path
from .mangling import mangle

__all__ = ["get_state", "update_state", "reducer", "ReducibleMeta", "Reducible"]


def get_state(obj):
    # type: (Any) -> dict[str, Any]
    """
    Get dictionary with an object's attribute values.
    Works with regular and slotted objects.

    :param obj: Object instance or class.
    :return: State dictionary.
    """
    state = {}  # type: dict[str, Any]

    # Get slotted values.
    if not isinstance(obj, type) and hasattr(type(obj), "__slots__"):
        cls = type(obj)
        for base in get_mro(cls):

            # Skip object.
            if base is object:
                continue

            # Get slot values.
            for slot in getattr(base, "__slots__", ()):

                # Skip weak reference slot.
                if slot == "__weakref__":
                    continue

                # Mangle slot if needed.
                slot = mangle(slot, base.__name__)

                # Skip if already has a value.
                if slot in state:
                    continue

                # Try to get value using the member descriptor.
                try:
                    state[slot] = base.__dict__[slot].__get__(obj, cls)
                except (KeyError, AttributeError):
                    pass

    # Get normal values.
    if hasattr(obj, "__dict__"):
        state.update(dict(obj.__dict__))

    return state


def update_state(obj, state_update):
    # type: (Any, Mapping[str, Any]) -> None
    """
    Update attribute values for an object.
    Works with regular and slotted objects.

    :param obj: Object instance or class.
    :param state_update: Dictionary with state updates.
    """
    remaining = set(state_update)  # type: set[str]

    # Set slotted attributes.
    if not isinstance(obj, type) and hasattr(type(obj), "__slots__"):
        cls = type(obj)
        for base in get_mro(cls):

            # Nothing left to do.
            if not remaining:
                return

            # Skip object.
            if base is object:
                continue

            # Set slot values.
            for slot in tuple(remaining):

                # Skip if not a slot in this base.
                member = base.__dict__.get(slot)
                if not isinstance(member, types.MemberDescriptorType):
                    continue

                # Set value using the member descriptor.
                member.__set__(obj, state_update[slot])
                remaining.remove(slot)

    # Set remaining attributes normally.
    for name in remaining:
        if isinstance(obj, type):
            type.__setattr__(obj, name, state_update[name])
        else:
            object.__setattr__(obj, name, state_update[name])


def _reducer(cls_or_path, state):
    if isinstance(cls_or_path, six.string_types):
        cls = import_path(cls_or_path)
    else:
        cls = cls_or_path
    self = cls.__new__(cls)
    update_state(self, state)
    return self


def _make_reducer(name, owner=None):
    # type: (str, Type | None) -> Callable
    script = """def {}(self):
    \"\"\"Reducer method that supports qualified name and slots for Python 2.7.\"\"\"
    cls = type(self)
    try:
        cls_or_path = get_path(cls)
    except ImportError:
        cls_or_path = cls
    return _reducer, (cls_or_path, get_state(self))
""".format(
        name
    )
    if owner is None:
        module = __name__
        owner_name = None
    else:
        module = owner.__module__
        owner_name = owner.__name__
    return make_function(
        name,
        script,
        globs={"get_path": get_path, "_reducer": _reducer, "get_state": get_state},
        filename=generate_unique_filename(name, module=module, owner_name=owner_name),
        module=module,
    )


reducer = _make_reducer("reducer")


class ReducibleMeta(type):
    """Metaclass that allows slotted classes to be pickled in Python 2.7."""

    if sys.version_info[0:2] < (3, 4):

        @staticmethod
        def __new__(mcs, name, bases, dct, **kwargs):
            cls = super(ReducibleMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)
            old_reducer = getattr(cls, "__reduce__", None)
            if old_reducer is None or old_reducer is object.__reduce__:
                type.__setattr__(cls, "__reduce__", _make_reducer("__reducer__", cls))
            return cls


class Reducible(six.with_metaclass(ReducibleMeta, object)):
    """
    Class that allows slotted classes to be pickled in Python 2.7.
    See `PEP 307 <https://peps.python.org/pep-0307/>`_.
    """

    __slots__ = ()



import six
import slotted
from tippo import GenericMeta

from .abstract_class import Abstract, AbstractMeta, abstract, is_abstract
from .default_dir import DefaultDir, DefaultDirMeta
from .explicit_hash import ExplicitHash, ExplicitHashMeta
from .implicit_hash import ImplicitHash, ImplicitHashMeta
from .init_subclass import InitSubclass, InitSubclassMeta
from .locked_class import LockedClass, LockedClassMeta, is_locked, set_locked, unlocked_context
from .namespace import Namespaced, NamespacedMeta
from .obj_state import Reducible, ReducibleMeta, get_state, update_state
from .qualname import Qualnamed, QualnamedMeta
from .runtime_final import RuntimeFinal, RuntimeFinalMeta, final
from .safe_not_equals import SafeNotEquals, SafeNotEqualsMeta
from .set_name import SetName, SetNameMeta

__all__ = [
    "CompatBaseMeta",
    "CompatBase",
    "BaseMeta",
    "Base",
    "SlottedBaseMeta",
    "SlottedBase",
    "abstract",
    "is_abstract",
    "final",
    "unlocked_context",
    "set_locked",
    "is_locked",
]


class CompatBaseMeta(
    InitSubclassMeta,
    DefaultDirMeta,
    ImplicitHashMeta,
    ReducibleMeta,
    QualnamedMeta,
    SafeNotEqualsMeta,
    SetNameMeta,
    AbstractMeta,
    GenericMeta,
):
    """Base metaclass for better compatibility amongst different Python versions."""


class CompatBase(
    six.with_metaclass(
        CompatBaseMeta,
        InitSubclass,
        DefaultDir,
        ImplicitHash,
        Reducible,
        Qualnamed,
        SafeNotEquals,
        SetName,
        Abstract,
    )
):
    """Base class for better compatibility amongst different Python versions."""

    __slots__ = ()


class BaseMeta(
    LockedClassMeta,
    ExplicitHashMeta,
    NamespacedMeta,
    RuntimeFinalMeta,
    CompatBaseMeta,
):
    """Base metaclass that adds extra features to the basic `type`."""


class Base(
    six.with_metaclass(
        BaseMeta,
        LockedClass,
        ExplicitHash,
        Namespaced,
        RuntimeFinal,
        CompatBase,
    )
):
    """Base class that adds extra features to the basic `object`."""

    __slots__ = ("__weakref__",)

    def __copy__(self):
        cls = type(self)
        new_self = cls.__new__(cls)
        update_state(new_self, get_state(self))
        return new_self


class SlottedBaseMeta(BaseMeta, slotted.SlottedABCGenericMeta):
    """Slotted base metaclass."""


class SlottedBase(six.with_metaclass(SlottedBaseMeta, Base, slotted.SlottedABC)):
    """Slotted base class."""

    __slots__ = ()



"""Dynamic code generation."""

import linecache

from tippo import Any, Callable

__all__ = ["compile_and_eval", "generate_unique_filename", "make_function"]


def compile_and_eval(script, globs=None, locs=None, filename=""):
    # type: (str, dict[str, Any] | None, dict[str, Any] | None, str) -> None
    """
    Evaluate the script with the given globals and locals.

    :param script: Script.
    :param globs: Globals.
    :param locs: Locals.
    :param filename: Filename.
    """
    bytecode = compile(script, filename, "exec")
    eval(bytecode, globs, locs)


def generate_unique_filename(obj_name, module=None, owner_name=None):
    # type: (str, str | None, str | None) -> str
    """
    Create a unique 'filename' suitable for a generated object.

    :param obj_name: Object name.
    :param module: Module name.
    :param owner_name: Name of the class that will own the method.
    :return: Filename.
    """
    if owner_name is not None:
        if module is None:
            error = "specified 'owner_name' but no 'module'"
            raise ValueError(error)
        return "<generated {}.{}.{}>".format(module, owner_name, obj_name)
    elif module is not None:
        return "<generated {}.{}>".format(module, obj_name)
    else:
        return "<generated {}>".format(obj_name)


def make_function(name, script, globs=None, filename=None, module=None):
    # type: (str, str, dict[str, Any] | None, str | None, str | None) -> Callable
    """
    Create a function with the given script.

    :param name: Function name.
    :param script: Script.
    :param filename: Filename.
    :param globs: Globals.
    :param module: Module name.
    :return: Function object.
    """

    # Add a fake linecache entry for debuggers.  # FIXME: infinite looping here when dup function is generated
    if filename:
        count = 1
        base_filename = complete_filename = filename  # type: str
        while True:
            linecache_tuple = (
                len(script),
                None,
                script.splitlines(True),
                complete_filename,
            )  # type: tuple[int, float | None, list[str], str]
            old_val = linecache.cache.setdefault(complete_filename, linecache_tuple)
            if old_val == linecache_tuple:
                break
            else:
                complete_filename = "{}-{}>".format(base_filename[:-1], count)
                count += 1
    else:
        complete_filename = ""

    # Compile and evaluate.
    locs = {}  # type: dict[str, Any]
    compile_and_eval(script, globs, locs, complete_filename)
    func = locs[name]

    # Set module name.
    if module is not None:
        object.__setattr__(func, "__module__", module)

    return func



import pytest  # noqa
import six

from basicco.qualname import Qualnamed, QualnamedMeta, QualnameError, qualname


@pytest.mark.parametrize("force_ast", (True, False))
def test_qualname(force_ast):

    # Get qualified name for regular classes.
    assert qualname(X, force_ast=force_ast) == "X"
    assert qualname(X.Y, force_ast=force_ast) == "X.Y"
    assert qualname(X.Y.Z, force_ast=force_ast) == "X.Y.Z"
    assert qualname(X.Y.Z.method, force_ast=force_ast) == "X.Y.Z.method"

    assert qualname(A, force_ast=force_ast) == "A"
    assert qualname(A.Y, force_ast=force_ast) == "A.Y"
    assert qualname(A.Y.C, force_ast=force_ast) == "A.Y.C"
    assert qualname(A.Y.C.method, force_ast=force_ast) == "A.Y.C.method"

    # AST parsing should fail for classes generated on the fly.
    with pytest.raises(QualnameError):
        print(qualname(type("Z", (object,), {"BLA": 1}), force_ast=True))

    # Built-ins.
    assert qualname(int, force_ast=force_ast) == "int"
    assert qualname(float, force_ast=force_ast) == "float"
    assert qualname(bool, force_ast=force_ast) == "bool"


def test_class():
    assert isinstance(Qualnamed, QualnamedMeta)

    assert "{}.QA.QY.QC.QD".format(__name__) in repr(QA.QY.QC.QD())


def test_metaclass():
    assert QX.__qualname__ == "QX"
    assert QX.QY.__qualname__ == "QX.QY"
    assert QX.QY.QZ.__qualname__ == "QX.QY.QZ"

    assert QA.__qualname__ == "QA"
    assert QA.QY.__qualname__ == "QA.QY"
    assert QA.QY.QC.__qualname__ == "QA.QY.QC"

    assert repr(QX.QY.QZ) == "<class '{}.QX.QY.QZ'>".format(__name__)
    assert repr(QA.QY.QC) == "<class '{}.QA.QY.QC'>".format(__name__)

    QX.QY.QZ.__qualname__ = "QX.QY.QZ.CUSTOM"
    QX.QY.__qualname__ = "QX.QY.CUSTOM"
    QX.__qualname__ = "QX.CUSTOM"

    QA.QY.QC.__qualname__ = "QA.QY.QC.CUSTOM"
    QA.QY.__qualname__ = "QA.QY.CUSTOM"
    QA.__qualname__ = "QA.CUSTOM"

    assert QX.__qualname__ == "QX.CUSTOM"
    assert QX.QY.__qualname__ == "QX.QY.CUSTOM"
    assert QX.QY.QZ.__qualname__ == "QX.QY.QZ.CUSTOM"

    assert QA.__qualname__ == "QA.CUSTOM"
    assert QA.QY.__qualname__ == "QA.QY.CUSTOM"
    assert QA.QY.QC.__qualname__ == "QA.QY.QC.CUSTOM"

    QX.QY.QZ.__qualname__ = "QX.QY.QZ"
    QX.QY.__qualname__ = "QX.QY"
    QX.__qualname__ = "QX"

    QA.QY.QC.__qualname__ = "QA.QY.QC"
    QA.QY.__qualname__ = "QA.QY"
    QA.__qualname__ = "QA"

    assert QX.__qualname__ == "QX"
    assert QX.QY.__qualname__ == "QX.QY"
    assert QX.QY.QZ.__qualname__ == "QX.QY.QZ"

    assert QA.__qualname__ == "QA"
    assert QA.QY.__qualname__ == "QA.QY"
    assert QA.QY.QC.__qualname__ == "QA.QY.QC"


class X(object):
    class Y(object):
        class Z(object):
            def method(self):
                pass


class A(object):
    class Y(object):
        class C(object):
            def method(self):
                pass


class QX(six.with_metaclass(QualnamedMeta, object)):
    class QY(six.with_metaclass(QualnamedMeta, object)):
        class QZ(six.with_metaclass(QualnamedMeta, object)):
            pass


class QA(six.with_metaclass(QualnamedMeta, object)):
    class QY(six.with_metaclass(QualnamedMeta, object)):
        class QC(six.with_metaclass(QualnamedMeta, object)):
            class QD(Qualnamed):
                pass


if __name__ == "__main__":
    pytest.main()



"""
Better matches Python 3.7+ behavior on Python 2.7 for
`abstract members <https://docs.python.org/3/library/abc.html>`_ while adding some additional features.
"""

import abc
import functools
import inspect
from abc import abstractmethod as abstract

import six
from tippo import Any, Set

from .dynamic_code import generate_unique_filename, make_function
from .get_mro import get_mro

__all__ = ["abstract", "is_abstract", "AbstractMeta", "Abstract"]

_ABSTRACT_CLASS_TAG = "__isabstractclass__"
_ABSTRACT_METHOD_TAG = "__isabstractmethod__"
_ABSTRACT_METHODS = "__abstractmethods__"


__abstract = abstract

try:
    _ABC = abc.ABC
except AttributeError:
    _ABC = object  # type: ignore


def _abstract(obj):
    """
    Decorate a class or member as abstract.

    :param obj: Class or member.
    :return: Decorated.
    """
    if inspect.isclass(obj):
        original_new = obj.__dict__.get("__new__")
        if isinstance(original_new, staticmethod):
            original_new = original_new.__func__

        script = "\n".join(
            (
                "def __new__(cls, *args, **kwargs):",
                "    if cls.__dict__.get(_ABSTRACT_CLASS_TAG, False):",
                "        abstract_members = getattr(cls, _ABSTRACT_METHODS, ())",
                "        if abstract_members:",
                "            error = 'can\\'t instantiate abstract class {!r} with abstract members {}'.format(",
                "                cls.__name__, ', '.join(repr(m) for m in sorted(abstract_members))",
                "            )",
                "        else:",
                "            error = 'can\\'t instantiate abstract class {!r}'.format(cls.__name__)",
                "        raise TypeError(error)",
                "    elif original_new is not None:",
                "        return original_new(cls, *args, **kwargs)",
                "    else:",
                "        return super(obj, cls).__new__(cls, *args, **kwargs)",
            )
        )
        func_name = "__new__"
        module = obj.__module__
        func = make_function(
            name=func_name,
            script=script,
            globs={
                "_ABSTRACT_CLASS_TAG": _ABSTRACT_CLASS_TAG,
                "_ABSTRACT_METHODS": _ABSTRACT_METHODS,
                "original_new": original_new,
                "obj": obj,
            },
            filename=generate_unique_filename(func_name, module=module, owner_name=obj.__name__),
            module=module,
        )
        func = functools.wraps(original_new or object.__new__)(func)
        type.__setattr__(obj, "__new__", staticmethod(func))
        type.__setattr__(obj, _ABSTRACT_CLASS_TAG, True)

        return obj

    else:
        return __abstract(obj)


globals()["abstract"] = _abstract  # trick IDEs/static type checkers


def is_abstract(obj):
    # type: (Any) -> bool
    """
    Tells whether a class or member is abstract.

    :param obj: Class or member.
    :return: True if abstract.
    """
    if inspect.isclass(obj):
        return obj.__dict__.get(_ABSTRACT_CLASS_TAG, False) or bool(getattr(obj, _ABSTRACT_METHODS, ()))
    else:
        return _is_abstract_member(obj)


def _is_abstract_member(obj):
    # type: (Any) -> bool
    """
    Tells whether a member is abstract.

    :param obj: Member.
    :return: True if abstract.
    """
    _is_abstract = False

    # Descriptor.
    if hasattr(obj, "__get__"):
        _is_abstract |= getattr(obj, _ABSTRACT_METHOD_TAG, False)

        # Has 'fget' getter (property-like).
        if hasattr(obj, "fget"):
            _is_abstract |= getattr(obj.fget, _ABSTRACT_METHOD_TAG, False)

    # Static or class method.
    if isinstance(obj, (staticmethod, classmethod)):
        _is_abstract |= getattr(obj.__func__, _ABSTRACT_METHOD_TAG, False)

    # Regular method.
    if callable(obj):
        _is_abstract |= getattr(obj, _ABSTRACT_METHOD_TAG, False)

    return _is_abstract


class AbstractMeta(abc.ABCMeta):
    """Metaclass that finds abstract members in properties and descriptors for both Python 2.7 and 3.7+."""

    def __init__(cls, name, bases, dct, **kwargs):
        super(AbstractMeta, cls).__init__(name, bases, dct, **kwargs)  # noqa
        cls.__gather_abstract_members()

    def __gather_abstract_members(cls):
        # type: () -> None

        # Iterate over MRO of the class.
        abstract_method_names = set()  # type: Set[str]
        for base in reversed(get_mro(cls)):

            # Find abstract members.
            for member_name, member in six.iteritems(base.__dict__):

                # Keep track.
                if _is_abstract_member(member):
                    abstract_method_names.add(member_name)
                else:
                    abstract_method_names.discard(member_name)

        # Update class information.
        type.__setattr__(cls, _ABSTRACT_METHODS, frozenset(abstract_method_names))

    def __setattr__(cls, name, value):
        super(AbstractMeta, cls).__setattr__(name, value)
        if name in cls.__dict__ and is_abstract(cls.__dict__[name]):
            cls.__gather_abstract_members()

    def __delattr__(cls, name):
        gather = False
        for base in get_mro(cls):
            if name in base.__dict__:
                gather = is_abstract(base.__dict__[name])
                break
        super(AbstractMeta, cls).__delattr__(name)
        if gather:
            cls.__gather_abstract_members()


class Abstract(six.with_metaclass(AbstractMeta, _ABC)):
    """Class that finds abstract members in properties and descriptors for both Python 2.7 and 3.7+."""

    __slots__ = ()



import sys

import pytest  # noqa

from basicco.set_name import SetName, SetNameMeta


class Attribute(object):
    def __set_name__(self, owner, name):
        assert name == "attribute_{}".format(owner.__name__.lower())
        self.owner = owner
        self.name = name


def test_meta():
    assert isinstance(SetName, SetNameMeta)


def test_set_name():
    class A(SetName):
        attribute_type_a = Attribute
        attribute_a = Attribute()

    class B(A):
        attribute_type_b = Attribute
        attribute_b = Attribute()

    class C(B):
        attribute_type_c = Attribute
        attribute_c = Attribute()

    assert C.attribute_a.name == "attribute_a"
    assert C.attribute_b.name == "attribute_b"
    assert C.attribute_c.name == "attribute_c"

    assert C.attribute_a.owner == A
    assert C.attribute_b.owner == B
    assert C.attribute_c.owner == C


def test_newer_python():
    if sys.version_info[:3] >= (3, 6):
        code = """
class A(SetName):
    attribute_type_a = Attribute
    attribute_a = Attribute()

class B(A):
    attribute_type_b = Attribute
    attribute_b = Attribute()

class C(B):
    attribute_type_c = Attribute
    attribute_c = Attribute()

assert C.attribute_a.name == "attribute_a"
assert C.attribute_b.name == "attribute_b"
assert C.attribute_c.name == "attribute_c"

assert C.attribute_a.owner == A
assert C.attribute_b.owner == B
assert C.attribute_c.owner == C
"""
        exec(code, globals(), globals())


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa
import six

from basicco.default_dir import DefaultDir, DefaultDirMeta


def test_class():
    assert isinstance(DefaultDir, DefaultDirMeta)


def test_metaclass():
    class Class(six.with_metaclass(DefaultDirMeta, object)):
        c = None
        d = None

        @property
        def prop_a(self):
            return None

    class SubClass(Class):
        __slots__ = ("a", "b")

        def __dir__(self):
            return super(SubClass, self).__dir__()

        @property
        def prop_b(self):
            return None

    obj = SubClass()
    assert set(dir(obj)).issuperset(
        {
            "__class__",
            "__delattr__",
            "__dict__",
            "__dir__",
            "__doc__",
            "__format__",
            "__getattribute__",
            "__hash__",
            "__init__",
            "__module__",
            "__new__",
            "__reduce__",
            "__reduce_ex__",
            "__repr__",
            "__setattr__",
            "__sizeof__",
            "__str__",
            "__subclasshook__",
            "__weakref__",
            "a",
            "b",
            "c",
            "d",
            "prop_a",
            "prop_b",
        }
    )


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa

from basicco import caller_module


def test_caller_module():
    def func():
        return caller_module.caller_module()

    assert func() == __name__


def test_auto_called_module():
    @caller_module.auto_caller_module("extra_paths", "cls_module")
    def func(extra_paths=(), cls_module=None, expects=((), None)):
        assert extra_paths == expects[0]
        assert cls_module == expects[1]

    func(expects=((__name__,), __name__))
    func(extra_paths=("foo", "bar"), expects=(("foo", "bar"), __name__))
    func(cls_module="foobar", expects=((__name__,), "foobar"))
    func(extra_paths=("foo", "bar"), cls_module="foobar", expects=(("foo", "bar"), "foobar"))


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa
from tippo import Any

from basicco.dynamic_code import compile_and_eval, generate_unique_filename, make_function


def test_generate_unique_filename():
    assert generate_unique_filename("my_func") == "<generated my_func>"
    assert generate_unique_filename("my_func", "module") == "<generated module.my_func>"
    assert generate_unique_filename("my_method", "module", "Class") == "<generated module.Class.my_method>"


def test_compile_and_eval():
    globs = {}  # type: dict[str, Any]
    compile_and_eval("foo = 40", globs)
    compile_and_eval("bar = 2", globs)
    compile_and_eval("foobar = foo + bar", globs)
    assert globs["foo"] == 40
    assert globs["bar"] == 2
    assert globs["foobar"] == 42


def test_make_function():
    name = "my_func"
    script = """
def my_func(foo=foo_default, bar=bar_default):
    return foo + bar
"""
    globs = {"foo_default": 1, "bar_default": 2}
    filename = generate_unique_filename("my_func", __name__)
    module = __name__
    my_func = make_function(name, script, globs, filename, module)

    assert my_func() == 3
    assert my_func(5, 5) == 10


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa
import six
import tippo

from basicco.explicit_hash import ExplicitHash, ExplicitHashMeta, set_to_none


def test_class():
    assert isinstance(ExplicitHash, ExplicitHashMeta)


def test_force_explicit_hash_declaration():
    class Class(six.with_metaclass(ExplicitHashMeta, object)):
        pass

    with pytest.raises(TypeError):

        class SubClass(Class):
            def __eq__(self, other):
                raise NotImplementedError()

        assert not SubClass

    class SubClass(Class):
        def __hash__(self):
            raise NotImplementedError()

        def __eq__(self, other):
            raise NotImplementedError()

    assert SubClass


def test_set_to_none():
    class Class(object):
        @set_to_none
        def __hash__(self):
            error = "not hashable"
            raise TypeError(error)

    assert not issubclass(Class, tippo.Hashable)
    assert not isinstance(Class(), tippo.Hashable)
    assert Class.__hash__ is None


if __name__ == "__main__":
    pytest.main()



"""
Python 2.7 compatible way of getting the qualified name.
Based on `wbolster/qualname <https://github.com/wbolster/qualname>`_.
For more information about the native Python 3 implementation, see
`PEP 3155 <https://peps.python.org/pep-03155/>`_.
"""

import ast
import inspect
import sys

import six
from tippo import Any, Callable

from .mangling import mangle

__all__ = ["QualnameError", "qualname", "QualnamedMeta", "Qualnamed"]


_cache = {}  # type: dict[str, dict[int, str]]


class QualnameError(Exception):
    """Raised when could not get the qualified name from AST parsing."""


def qualname(obj, fallback=None, force_ast=False, use_existing=True):
    # type: (Callable, str | None, bool, bool) -> str
    """
    Try to find out the qualified name for a class or function.
    This function uses AST parsing in Python 2.7 to try to replicate the qualified name functionality available in
    Python 3.3+.

    :param obj: Function or class.
    :param fallback: Value to return if we couldn't get qualified name. Will error if None.
    :param force_ast: Force use AST parser to get qualified name from code.
    :param use_existing: Whether to use existing '__qualname__' attribute if available.
    :return: Qualified name (or None if not available).
    :raises TypeError: Provided object is not a function or a class.
    :raises QualnameError: Could not get qualified name from AST parsing.
    """

    # Not available for non-callables.
    if not callable(obj) or not hasattr(obj, "__name__") or not hasattr(obj, "__module__"):
        error = "can't determine qualified name for instances of {!r}".format(type(obj).__name__)
        raise TypeError(error)
    obj_name = obj.__name__
    obj_module = obj.__module__  # type: ignore

    # Native qualified name or manually defined.
    if not force_ast:
        try:
            if not use_existing:
                error = "{!r} object has '__qualname__' attribute, but 'use_existing' was declared as False"
                raise QualnameError(error)
            return getattr(obj, "__qualname__")
        except QualnameError:
            if fallback is not None:
                return fallback
            else:
                raise
        except AttributeError:
            pass

    # Try to match the root of the module in case the object is not nested.
    try:
        module = __import__(obj_module, fromlist=[obj_name])
    except ImportError as e:
        error = "couldn't import module {!r} for {!r}; {}".format(obj_module, obj_name, e)
        exc = QualnameError(error)
        six.raise_from(exc, None)
        raise exc
    if getattr(module, obj_name, None) is obj:
        return obj_name

    # Get source file name.
    try:
        filename = inspect.getsourcefile(obj)
        if filename is None:
            raise TypeError()
    except TypeError:
        if fallback is None:
            error = "couldn't find source code filename for {!r}".format(obj_name)
            exc = QualnameError(error)
            six.raise_from(exc, None)
            raise exc
        else:
            return fallback

    # Get line number.
    if inspect.isclass(obj):
        try:
            _, lineno = inspect.getsourcelines(obj)
        except (OSError, IOError):
            if fallback is None:
                error = "source code for class {!r} could not be retrieved".format(obj_name)
                exc = QualnameError(error)
                six.raise_from(exc, None)
                raise exc
            else:
                return fallback
    elif inspect.isfunction(obj) or inspect.ismethod(obj):
        if hasattr(obj, "im_func"):
            # Extract function from unbound method (Python 2).
            obj = obj.im_func  # type: ignore
        try:
            code = obj.__code__
        except AttributeError:
            code = obj.func_code  # type: ignore
        lineno = code.co_firstlineno
    else:
        if fallback is None:
            error = "source code could not be retrieved for {!r}".format(obj_name)
            raise QualnameError(error)
        else:
            return fallback

    # Parse the source file to figure out what the qualified name should be.
    if filename in _cache:
        qualified_names = _cache[filename]
    else:
        try:
            with open(filename, "r") as fp:
                source = fp.read()
        except (OSError, IOError) as e:
            if fallback is None:
                error = "could not read source code at {!r}; {}".format(filename, e)
                exc = QualnameError(error)
                six.raise_from(exc, None)
                raise exc
            else:
                return fallback
        node = ast.parse(source, filename)
        visitor = _Visitor()
        visitor.visit(node)
        qualified_names = _cache[filename] = visitor.qualified_names

    # Sort list of line numbers based on the one found via inspection.
    if qualified_names:
        linenos = sorted(
            qualified_names,
            reverse=True,
            key=lambda k: (k - lineno if k >= lineno else max(qualified_names) + (lineno - k)),
        )
    else:
        linenos = []

    # Iterate over possible line numbers.
    while linenos:
        current_lineno = linenos.pop()

        # Get qualified name from parsing results.
        qualified_name = qualified_names.get(current_lineno, None)
        if qualified_name is None:
            if fallback is None:
                error = "qualified name could not be retrieved from source code for {!r}".format(obj_name)
                raise QualnameError(error)
            else:
                return fallback

        # Verify qualified name.
        if qualified_name and "<locals>" not in qualified_name:
            qualified_name_parts = qualified_name.split(".")
            test_obj = module
            while qualified_name_parts:
                part = qualified_name_parts.pop(0)
                try:
                    test_obj = getattr(test_obj, part)
                except AttributeError:
                    break
            else:

                # Extract function from unbound method (Python 2).
                if hasattr(test_obj, "im_func"):
                    test_obj = test_obj.im_func  # type: ignore

                # Return only confirmed to be the same exact object.
                if test_obj is obj and test_obj:
                    return qualified_name

    # No way to reliably retrieve qualified name.
    if fallback is None:
        error = "qualified name could not be retrieved from source code for {!r}".format(obj.__name__)
        raise QualnameError(error)
    else:
        return fallback


class QualnamedMeta(type):
    """Metaclass that implements `__qualname__` for Python 2.7."""

    if sys.version_info[:2] < (3, 4):
        assert not hasattr(object, "__qualname__")

        def __repr__(cls):
            # type: () -> str
            """
            Get representation using the class' full name if possible.

            :return: Class representation.
            """
            return "<class '{}.{}'>".format(cls.__module__, cls.__qualname__)

        def __getattr__(cls, name):
            # type: (str) -> Any
            attr = mangle("__qualname", cls.__name__)
            if name == attr:
                qualified_name = qualname(cls, fallback=cls.__name__, force_ast=True, use_existing=False)
                type.__setattr__(cls, attr, qualified_name)
                return qualified_name
            try:
                return super(QualnamedMeta, cls).__getattr__(name)  # type: ignore  # noqa
            except AttributeError:
                pass
            error = "class {!r} has no attribute {!r}".format(cls.__name__, name)
            raise AttributeError(error)

        @property
        def __qualname__(cls):
            # type: () -> str
            """Qualified name."""
            attr = mangle("__qualname", cls.__name__)
            return getattr(cls, attr)

        @__qualname__.setter
        def __qualname__(cls, value):
            # type: (str) -> None
            attr = mangle("__qualname", cls.__name__)
            setattr(cls, attr, value)


class Qualnamed(six.with_metaclass(QualnamedMeta, object)):
    """Class that implements `__qualname__` for Python 2.7."""

    __slots__ = ()

    if sys.version_info[:2] < (3, 4):

        def __repr__(self):
            # type: () -> str
            """
            Get representation using the class' full name if possible.

            :return: Representation.
            """
            return "<{}.{} at {}>".format(type(self).__module__, type(self).__qualname__, hex(id(self)))


class _Visitor(ast.NodeVisitor):
    def __init__(self):
        self.stack = []
        self.qualified_names = {}

    def store_qualified_name(self, lineno):
        qn = ".".join(n for n in self.stack)
        self.qualified_names[lineno] = qn

    def visit_FunctionDef(self, node):
        self.stack.append(node.name)
        self.store_qualified_name(node.lineno)
        self.stack.append("<locals>")
        self.generic_visit(node)
        self.stack.pop()
        self.stack.pop()

    def visit_ClassDef(self, node):
        self.stack.append(node.name)
        self.store_qualified_name(node.lineno)
        self.generic_visit(node)
        self.stack.pop()



import pytest  # noqa
from tippo import GenericMeta

import basicco


def test_bases():
    assert basicco.CompatBaseMeta
    assert basicco.CompatBase
    assert basicco.CompatBase()

    assert issubclass(basicco.CompatBaseMeta, GenericMeta)

    assert basicco.BaseMeta
    assert basicco.Base
    assert basicco.Base()
    assert "__weakref__" in basicco.Base.__slots__

    assert basicco.SlottedBaseMeta
    assert basicco.SlottedBase
    assert basicco.SlottedBase()


if __name__ == "__main__":
    pytest.main()



"""Backport of the functionality of `__set_name__` from PEP 487 to Python 2.7."""

import sys

import six

__all__ = ["SetNameMeta", "SetName"]


class SetNameMeta(type):
    """Metaclass that backports the functionality of `__set_name__` from PEP 487 to Python 2.7."""

    if sys.version_info[:3] < (3, 6):

        @staticmethod
        def __new__(mcs, name, bases, dct, **kwargs):
            cls = super(SetNameMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)
            for member_name, member in six.iteritems(dct):
                if hasattr(member, "__set_name__") and not isinstance(member, type):
                    member.__set_name__(cls, member_name)
            return cls


class SetName(six.with_metaclass(SetNameMeta, object)):
    """Class that backports the functionality of `__set_name__` from PEP 487 to Python 2.7."""

    __slots__ = ()



"""Weak Reference-like object that supports pickling."""

import copy
import threading
import weakref

from tippo import Any, Generic, Type, TypeVar

from .recursive_repr import recursive_repr

__all__ = ["WeakReference", "UniqueHashWeakReference"]


_T = TypeVar("_T")

_DEAD_REF = weakref.ref(type("Dead", (object,), {"__slots__": ("__weakref__",)})())

_cache = {}  # type: dict[Type["WeakReference"], weakref.WeakValueDictionary[int, "WeakReference"]]

_lock = threading.RLock()


def _reduce_dead(cls):
    self = super(WeakReference, cls).__new__(cls)  # type: ignore
    self.__init__(None)  # type: ignore
    return self


class WeakReference(Generic[_T]):
    """Weak Reference-like object that supports pickling."""

    __slots__ = ("__weakref__", "__ref")

    @staticmethod
    def __new__(cls, obj=None):
        # type: (Type[WeakReference[_T]], _T | None) -> WeakReference[_T]
        with _lock:
            cache = _cache.setdefault(cls, weakref.WeakValueDictionary())
            if obj is None:
                obj_ref = _DEAD_REF
            else:
                obj_ref = weakref.ref(obj)

            obj_ref_id = id(obj_ref)
            try:
                self = cache[obj_ref_id]
            except KeyError:
                pass
            else:
                try:
                    if self.__ref is obj_ref:
                        return self
                except ReferenceError:
                    pass

            self = cache[obj_ref_id] = super(WeakReference, cls).__new__(cls)
            return self

    def __init__(self, obj=None):
        # type: (_T | None) -> None
        """
        :param obj: Object to reference.
        """
        if obj is None:
            self.__ref = _DEAD_REF
        else:
            self.__ref = weakref.ref(obj)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        self_id = id(self)
        try:
            dup = memo[self_id]
        except KeyError:
            cls = type(self)
            obj = self()
            if self is type(self)():
                assert obj is None
                dup = memo[self_id] = self
            elif obj is None:
                dup = memo[self_id] = _reduce_dead(cls)
            else:
                dup = memo[self_id] = cls(copy.deepcopy(obj, memo))
        return dup

    def __hash__(self):
        return hash(self.__ref)

    def __eq__(self, other):
        # type: (Any) -> bool
        return self is other

    def __ne__(self, other):
        # type: (Any) -> bool
        return not self.__eq__(other)

    @recursive_repr
    def __repr__(self):
        obj = self()
        return "<{} object at {}; {}>".format(
            type(self).__name__,
            hex(id(self)),
            "dead"
            if obj is None
            else "to '{}' at {}".format(
                type(obj).__name__,
                hex(id(obj)),
            ),
        )

    def __str__(self):
        return self.__repr__()

    def __call__(self):
        # type: () -> _T | None
        """
        Get strong reference to the object or None if no longer alive.

        :return: Strong reference or `None`.
        """
        return self.__ref()

    def __reduce__(self):
        cls = type(self)
        obj = self()
        if self is type(self)():
            assert obj is None
            return cls, (None,)
        if obj is None:
            return _reduce_dead, (cls,)
        return cls, (obj,)


class UniqueHashWeakReference(WeakReference[_T]):
    """Weak Reference-like object that supports pickling and has a unique id-based hash."""

    __slots__ = ()

    def __hash__(self):
        return object.__hash__(self)



import pytest  # noqa
import six

from basicco.locked_class import LockedClass, LockedClassMeta, is_locked, set_locked, unlocked_context


def test_class():
    assert isinstance(LockedClass, LockedClassMeta)


def test_metaclass():
    class MyMeta(LockedClassMeta):
        @staticmethod
        def __new__(mcs, name, bases, dct, **kwargs):
            cls = super(MyMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)

            assert not is_locked(cls)
            cls.bar = "foo"

            return cls

        def __init__(cls, name, bases, dct, **kwargs):
            assert not is_locked(cls)
            cls.foobar = "foobar"

            super(MyMeta, cls).__init__(name, bases, dct, **kwargs)

            assert is_locked(cls)
            with pytest.raises(AttributeError):
                cls.foo = "bar"

    class Class(six.with_metaclass(MyMeta, object)):
        pass

    assert is_locked(Class)

    with pytest.raises(AttributeError):
        Class.foo = "bar"

    Class._foo = "bar"
    Class._Class__foo = "bar"
    Class.__foo__ = "bar"


def test_set_locked():
    class Class(six.with_metaclass(LockedClassMeta, object)):
        pass

    with pytest.raises(AttributeError):
        Class.foo = "bar"

    set_locked(Class, False)
    Class.foo = "bar"

    set_locked(Class, True)
    with pytest.raises(AttributeError):
        Class.foo = "bar"


def test_unlocked_context():
    class Class(six.with_metaclass(LockedClassMeta, object)):
        pass

    with pytest.raises(AttributeError):
        Class.foo = "bar"

    with unlocked_context(Class):
        Class.foo = "bar"

    with pytest.raises(AttributeError):
        Class.foo = "bar"


if __name__ == "__main__":
    pytest.main()



"""
Backport of Python 3's implementation of
`object.__dir__ <https://docs.python.org/3/reference/datamodel.html#object.__dir__>`_ for Python 2.7.
"""

import inspect

import six

from .dynamic_code import generate_unique_filename, make_function

__all__ = ["DefaultDirMeta", "DefaultDir"]


class DefaultDirMeta(type):
    """Metaclass that backports the base implementation of `__dir__` for Python 2.7."""

    if not hasattr(object, "__dir__"):

        @staticmethod
        def __new__(mcs, name, bases, dct, **kwargs):
            cls = super(DefaultDirMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)
            if not hasattr(cls, "__dir__"):
                script = "\n".join(
                    (
                        "def __dir__(self):",
                        "    cls = type(self)",
                        "    members = set()",
                        '    if hasattr(self, "__dict__"):',
                        "        members.update(self.__dict__)",
                        "    for base in inspect.getmro(cls):",
                        "        members.update(base.__dict__)",
                        "    return sorted(members)",
                    )
                )
                func_name = "__dir__"
                module = dct.get("__module__", None)
                func = make_function(
                    name=func_name,
                    script=script,
                    globs={"inspect": inspect},
                    filename=generate_unique_filename(func_name, module=module, owner_name=name),
                    module=module,
                )
                type.__setattr__(cls, func_name, func)
            return cls


class DefaultDir(six.with_metaclass(DefaultDirMeta, object)):
    """Class that backports the base implementation of `__dir__` for Python 2.7."""

    __slots__ = ()



"""
Backport of the default Python 3 behavior of `__ne__` behavior for Python 2.7.
`Read more <https://docs.python.org/3.0/whatsnew/3.0.html#operators-and-special-methods>`_.
"""

import functools
import sys

import six

from .dynamic_code import generate_unique_filename, make_function

__all__ = ["SafeNotEqualsMeta", "SafeNotEquals"]


class SafeNotEqualsMeta(type):
    """Metaclass that backports the default Python 3 behavior of `__ne__` behavior for Python 2.7."""

    if sys.version_info[:1] < (3,):

        @staticmethod
        def __new__(mcs, name, bases, dct, **kwargs):
            cls = super(SafeNotEqualsMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)
            if cls.__ne__ is object.__ne__:
                script = "def __ne__(self, other):\n    return not (self == other)"
                func_name = "__ne__"
                module = dct.get("__module__", None)
                func = make_function(
                    name=func_name,
                    script=script,
                    globs={},
                    filename=generate_unique_filename(func_name, module=module, owner_name=name),
                    module=module,
                )
                func = functools.wraps(object.__ne__)(func)
                type.__setattr__(cls, func_name, func)
            return cls


class SafeNotEquals(six.with_metaclass(SafeNotEqualsMeta, object)):
    """Class that backports the default Python 3 behavior of `__ne__` behavior for Python 2.7."""

    __slots__ = ()



import pytest
import six

from basicco.abstract_class import Abstract, AbstractMeta, abstract, is_abstract


def test_class():
    assert isinstance(Abstract, AbstractMeta)


def test_abstract_class():
    @abstract
    class Class(six.with_metaclass(AbstractMeta, object)):
        pass

    with pytest.raises(TypeError):
        Class()

    class SubClass(Class):
        pass

    assert SubClass()


def test_super_new():
    class Class(six.with_metaclass(AbstractMeta, object)):
        @staticmethod
        def __new__(cls):
            self = super(Class, cls).__new__(cls)
            self.new_called = True
            return self

    obj = Class()
    assert obj.new_called is True

    class SubClass(Class):
        pass

    obj = SubClass()
    assert obj.new_called is True


def test_abstract_method():
    def dummy_decorator(func):
        return func

    class PropertyLikeDescriptor(object):
        def __init__(self, fget):
            self.fget = fget

        def __get__(self, instance, owner):
            return 3

    decorators = [
        dummy_decorator,
        PropertyLikeDescriptor,
        classmethod,
        staticmethod,
        property,
    ]

    # Different kinds of decorated members should be recognized as abstract.
    for decorator in decorators:

        class Class(six.with_metaclass(AbstractMeta, object)):
            @decorator
            @abstract
            def method(self):
                pass

        assert is_abstract(Class.__dict__["method"])

        abstract_methods = getattr(Class, "__abstractmethods__")
        assert "method" in abstract_methods

        Class.new_method = decorator(abstract(lambda _: None))
        abstract_methods = getattr(Class, "__abstractmethods__")
        assert "new_method" in abstract_methods

        del Class.new_method
        abstract_methods = getattr(Class, "__abstractmethods__")
        assert "new_method" not in abstract_methods

        with pytest.raises(TypeError):
            Class()


def test_descriptor():
    class Descriptor(object):
        def __init__(self, func):
            self.func = func

        def __get__(self, instance, owner):
            return 3

    class Class(six.with_metaclass(AbstractMeta, object)):
        @abstract
        @Descriptor
        def method(self):
            pass

    # Decorated descriptor object should be recognized as abstract.
    assert "method" in getattr(Class, "__abstractmethods__")


if __name__ == "__main__":
    pytest.main()



"""Prevents changing public class attributes."""

import contextlib

import six
from tippo import Any, Iterator, Type

from .mangling import mangle
from .type_checking import assert_is_instance

__all__ = ["is_locked", "set_locked", "unlocked_context", "LockedClassMeta", "LockedClass"]


def is_locked(cls):
    # type: (Type) -> bool
    """
    Tell whether public attributes of a class are locked or not.

    :param cls: Class.
    :return: True if locked.
    """
    locked_attr = mangle("__locked", cls.__name__)
    try:
        locked = getattr(cls, locked_attr)
    except AttributeError:
        assert_is_instance(cls, LockedClassMeta)
        return True
    else:
        return locked


def set_locked(cls, locked):
    # type: (Type[LockedClass] | LockedClassMeta, bool) -> None
    """
    Set class locked state.

    :param cls: Class.
    :param locked: Locked state.
    :raise TypeCheckError: Invalid class type.
    """
    locked_attr = mangle("__locked", cls.__name__)
    if not hasattr(cls, locked_attr):
        assert_is_instance(cls, LockedClassMeta)
    elif not locked:
        type.__setattr__(cls, locked_attr, False)
    else:
        type.__delattr__(cls, locked_attr)


@contextlib.contextmanager
def unlocked_context(cls):
    # type: (Type[LockedClass] | LockedClassMeta) -> Iterator
    """
    Unlocked class context manager.

    :param cls: Class.
    :raise TypeCheckError: Invalid class type.
    """
    before = is_locked(cls)
    if before:
        set_locked(cls, False)
    try:
        yield
    finally:
        if before:
            set_locked(cls, True)


class LockedClassMeta(type):
    """Metaclass that prevents changing public class attributes."""

    def __new__(mcs, name, bases, dct, **kwargs):
        dct = dict(dct)
        dct[mangle("__locked", name)] = False
        return super(LockedClassMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)

    def __init__(cls, name, bases, dct, **kwargs):
        super(LockedClassMeta, cls).__init__(name, bases, dct, **kwargs)
        if not is_locked(cls):
            set_locked(cls, True)

    def __getattr__(cls, name):
        # type: (str) -> Any
        if name == mangle("__locked", cls.__name__):
            return True
        try:
            return super(LockedClassMeta, cls).__getattr__(name)  # type: ignore  # noqa
        except AttributeError:
            pass
        error = "class {!r} has no attribute {!r}".format(cls.__name__, name)
        raise AttributeError(error)

    def __setattr__(cls, name, value):
        """Prevent setting public class attributes."""
        if is_locked(cls) and not name.startswith("_"):
            error = "can't set read-only class attribute {!r}".format(name)
            raise AttributeError(error)
        super(LockedClassMeta, cls).__setattr__(name, value)

    def __delattr__(cls, name):
        """Prevent deleting class attributes."""
        if is_locked(cls) and not name.startswith("_"):
            error = "can't delete read-only class attribute {!r}".format(name)
            raise AttributeError(error)
        super(LockedClassMeta, cls).__delattr__(name)


class LockedClass(six.with_metaclass(LockedClassMeta, object)):
    """Class that prevents changing public class attributes."""

    __slots__ = ()



"""Metaclass that forces `__hash__` to be declared when `__eq__` is declared."""

import six
from tippo import Callable, TypeVar, cast

__all__ = ["ExplicitHashMeta", "ExplicitHash", "set_to_none"]


T = TypeVar("T")


class ExplicitHashMeta(type):
    """Metaclass that forces `__hash__` to be declared when `__eq__` is declared."""

    @staticmethod
    def __new__(mcs, name, bases, dct, **kwargs):
        if "__eq__" in dct and "__hash__" not in dct:
            error = "declared '__eq__' in {!r} but didn't declare '__hash__'".format(name)
            raise TypeError(error)
        return super(ExplicitHashMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)


class ExplicitHash(six.with_metaclass(ExplicitHashMeta, object)):
    """Class that forces `__hash__` to be declared when `__eq__` is declared."""

    __slots__ = ()


def set_to_none(_hash_method):
    # type: (Callable[[T], int]) -> Callable[[T], int]
    """
    Decorates a '__hash__' method, so it gets set to None for non-hashable classes.

    :param _hash_method: Hash method function.
    :return: None
    """
    return cast(Callable[[T], int], None)



"""Runtime type checking with support for import paths and type hints."""

import itertools

import six
import typing_inspect  # type: ignore
from tippo import Any, Callable, ForwardRef, Iterable, Mapping, Type, TypeVar, cast, get_args, get_origin, get_typing

from .import_path import get_name, import_path

__all__ = [
    "TEXT_TYPES",
    "TypeCheckError",
    "type_names",
    "format_types",
    "import_types",
    "is_instance",
    "is_subclass",
    "is_iterable",
    "assert_is_instance",
    "assert_is_subclass",
    "assert_is_callable",
    "assert_is_iterable",
]


TEXT_TYPES = tuple({str, six.text_type})  # type: tuple[Type, ...]

_T = TypeVar("_T")


class TypeCheckError(Exception):
    """Raised when failed to assert type check."""


def _check_literal(obj, literal, type_depth, *args):
    return any(_check(obj, type(v), type_depth, *args) if type_depth else obj == v for v in get_args(literal))


def _check_union(obj, union, *args):
    return any(_check(obj, t, *args) for t in get_args(union))


def _check_type(obj, typ, type_depth, *args):
    if not isinstance(obj, type):
        return False

    type_args = get_args(typ)
    if not type_args:
        return True

    assert len(type_args) == 1
    value_type = type_args[0]
    type_depth += 1

    return _check(obj, value_type, type_depth, *args)


def _check_tuple(obj, typed_tuple, type_depth, instance, *args):
    if type_depth or not instance:
        return _check(obj, tuple, type_depth, instance, *args)

    if not isinstance(obj, tuple):
        return False

    tuple_args = get_args(typed_tuple)
    if not tuple_args:
        return True

    if tuple_args[-1] == Ellipsis:
        if len(tuple_args) == 1:
            return True
        typ = tuple_args[0]
        return all(_check(v, typ, type_depth, instance, *args) for v in obj)

    if len(obj) != len(tuple_args):
        return False

    return all(_check(v, t, type_depth, instance, *args) for v, t in zip(obj, tuple_args))


def _check_mapping(obj, mapping, type_depth, instance, typing, *args):
    origin = get_origin(mapping)
    if type_depth or not instance:
        return _check(obj, origin, type_depth, instance, False, *args)

    if not isinstance(obj, origin):  # type: ignore
        return False

    mapping_args = get_args(mapping)
    if not mapping_args:
        return True

    assert len(mapping_args) == 2
    key_type, value_type = mapping_args
    for key, value in six.iteritems(obj):
        if not _check(key, key_type, type_depth, instance, typing, *args):
            return False
        if not _check(value, value_type, type_depth, instance, typing, *args):
            return False

    return True


def _check_iterable(obj, iterable, type_depth, instance, typing, *args):
    origin = get_origin(iterable)
    if type_depth or not instance:
        return _check(obj, origin, type_depth, instance, False, *args)

    if not isinstance(obj, origin):  # type: ignore
        return False

    iterable_args = get_args(iterable)
    if not iterable_args:
        return True

    assert len(iterable_args) == 1
    value_type = iterable_args[0]
    for value in obj:
        if not _check(value, value_type, type_depth, instance, typing, *args):
            return False

    return True


def _check_forward_ref(
    obj,
    forward_ref,
    type_depth,
    instance,
    typing,
    subtypes,
    extra_paths,
    builtin_paths,
    generic,
):
    typ = import_path(
        forward_ref.__forward_arg__,
        extra_paths=extra_paths,
        builtin_paths=builtin_paths,
        generic=generic,
    )
    return _check(
        obj,
        typ,
        type_depth,
        instance,
        typing,
        subtypes,
        extra_paths,
        builtin_paths,
        generic,
    )


def _check_typing(
    obj,
    typ,
    type_depth,
    instance,
    typing,
    subtypes,
    extra_paths,
    builtin_paths,
    generic,
):
    args = (
        type_depth,
        instance,
        typing,
        subtypes,
        extra_paths,
        builtin_paths,
        generic,
    )

    # Any.
    if typ is Any:
        return True

    # Literal.
    if typing_inspect.is_literal_type(typ):
        return _check_literal(obj, typ, *args)

    # Union.
    if typing_inspect.is_union_type(typ):
        return _check_union(obj, typ, *args)

    # Type.
    if get_typing(get_origin(typ)) is Type:
        return _check_type(obj, typ, *args)

    # Tuple.
    if typing_inspect.is_tuple_type(typ):
        return _check_tuple(obj, typ, *args)

    # Mapping.
    try:
        type_is_mapping = issubclass(get_typing(get_origin(typ)), Mapping)
    except TypeError:
        pass
    else:
        if type_is_mapping:
            return _check_mapping(obj, typ, *args)

    # Iterable.
    try:
        type_is_iterable = issubclass(get_typing(get_origin(typ)), Iterable)
    except TypeError:
        pass
    else:
        if type_is_iterable:
            return _check_iterable(obj, typ, *args)

    # Forward reference.
    if isinstance(typ, ForwardRef):
        return _check_forward_ref(obj, typ, *args)

    # Not typing.
    return _check(obj, typ, args[0], args[1], False, *args[3:])


def _check(
    obj,
    typ,
    type_depth,
    instance,
    typing,
    subtypes,
    extra_paths,
    builtin_paths,
    generic,
):

    # Import lazy path.
    if isinstance(typ, six.string_types):
        typ = import_path(
            typ,
            extra_paths=extra_paths,
            builtin_paths=builtin_paths,
            generic=generic,
        )

    # Convert None to NoneType.
    if typ is None:
        typ = type(None)

    # Typing check.
    if typing:
        typing_name = get_name(typ)
        if typing_name is not None:
            return _check_typing(
                obj,
                typ,
                type_depth,
                instance,
                typing,
                subtypes,
                extra_paths,
                builtin_paths,
                generic,
            )

    # Invalid type.
    if not isinstance(typ, type) and not hasattr(typ, "__subclasscheck__"):
        error = "{!r} object is not a valid type".format(type(typ).__name__)
        raise TypeError(error)

    # Resolve type depth.
    for i in range(type_depth - 1 if instance else type_depth):
        typ = type(typ)

    # Everything is an instance and subclass of object.
    if typ is object:
        return True

    # Switch to a subclass comparison when type depth is greater than 0.
    if type_depth > 0:
        if not isinstance(obj, type):
            return False
        instance = False

    # Instance checks.
    if instance:
        if subtypes:
            return isinstance(obj, typ)
        else:
            return type(obj) is typ

    # Subclass checks.
    elif not isinstance(obj, type):
        error = "{!r} object is not a class".format(type(obj).__name__)
        raise TypeError(error)
    elif subtypes:
        return issubclass(obj, typ)
    else:
        return obj is typ


def format_types(types):
    # type: (Type| str | None | Iterable[Type| str | None]) -> tuple[Type | str, ...]
    """
    Format types into a tuple.

    :param types: Type(s)/Import Path(s).
    :return: Tuple of Types/Import Paths.
    """
    if types is None:
        return (type(None),)
    elif typing_inspect.is_union_type(types):
        return tuple(itertools.chain.from_iterable(format_types(t) for t in get_args(types)))
    elif isinstance(types, type) or isinstance(types, six.string_types) or get_name(types) is not None:
        return (types,)  # type: ignore
    elif isinstance(types, Iterable):  # type: ignore
        return tuple(itertools.chain.from_iterable(format_types(t) for t in types))
    else:
        error = "{!r} object is not a valid type".format(type(types).__name__)
        raise TypeError(error)


def type_names(types):
    # type: (Type | str | None | Iterable[Type | str | None]) -> tuple[str, ...]
    """
    Get type names without importing from paths. Can be used for user-feedback purposes.

    :param types: Types.
    :return: Type names.
    """
    type_names_ = []
    for typ in format_types(types):
        if isinstance(typ, six.string_types):
            type_names_.append(typ.split(".")[-1])
        elif isinstance(typ, type):
            type_names_.append(typ.__name__)
        elif get_name(typ) is not None:
            type_names_.append(repr(typ))
        else:
            type_names_.append(type(typ).__name__)
    return tuple(type_names_)


def import_types(
    types,  # type: Type | str | None | Iterable[Type | str | None]
    extra_paths=(),  # type: Iterable[str]
    builtin_paths=None,  # type: Iterable[str] | None
    generic=True,  # type: bool
):
    # type: (...) -> tuple[Type, ...]
    """
    Import types from lazy import paths.

    :param types: Types.
    :param extra_paths: Extra module paths in fallback order.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :return: Imported types.
    """
    imported_types = tuple(
        import_path(t, extra_paths=extra_paths, builtin_paths=builtin_paths, generic=generic)
        if isinstance(t, six.string_types)
        else t
        for t in format_types(types)
    )

    # Expand all text types.
    if len(TEXT_TYPES) > 1:
        text_types = set(TEXT_TYPES)
        if text_types.intersection(imported_types):
            imported_types += tuple(text_types.difference(imported_types))

    # Expand all integer types.
    if len(six.integer_types) > 1:
        integer_types = set(six.integer_types)
        if integer_types.intersection(imported_types):
            imported_types += tuple(integer_types.difference(imported_types))

    return cast("tuple[Type, ...]", format_types(imported_types))


def is_instance(
    obj,  # type: Any
    types,  # type: Type | str | None | Iterable[Type | str | None]
    subtypes=True,  # type: bool
    extra_paths=(),  # type: Iterable[str]
    builtin_paths=None,  # type: Iterable[str] | None
    generic=True,  # type: bool
    typing=True,  # type: bool
):
    # type: (...) -> bool
    """
    Get whether object is an instance of the provided types.

    :param obj: Object.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param extra_paths: Extra module paths in fallback order.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :param typing: Whether to check against typing.
    :return: True if it is an instance.
    """
    imported_types = import_types(types, extra_paths=extra_paths, builtin_paths=builtin_paths, generic=generic)
    return any(
        _check(
            obj=obj,
            typ=t,
            type_depth=0,
            instance=True,
            typing=typing,
            subtypes=subtypes,
            extra_paths=extra_paths,
            builtin_paths=builtin_paths,
            generic=generic,
        )
        for t in imported_types
    )


def is_subclass(
    cls,  # type: Type
    types,  # type: Type | str | None | Iterable[Type | str | None]
    subtypes=True,  # type: bool
    extra_paths=(),  # type: Iterable[str]
    builtin_paths=None,  # type: Iterable[str] | None
    generic=True,  # type: bool
    typing=True,  # type: bool
):
    # type: (...) -> bool
    """
    Get whether class is a subclass of the provided types.

    :param cls: Class.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param extra_paths: Extra module paths in fallback order.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :param typing: Whether to check against typing.
    :return: True if it is a subclass.
    :raises TypeError: Did not provide a class.
    """
    if not isinstance(cls, type):
        error = "is_subclass() arg 1 must be a class"
        raise TypeError(error)
    imported_types = import_types(types, extra_paths=extra_paths, builtin_paths=builtin_paths, generic=generic)
    return any(
        _check(
            obj=cls,
            typ=t,
            type_depth=0,
            instance=False,
            typing=typing,
            subtypes=subtypes,
            extra_paths=extra_paths,
            builtin_paths=builtin_paths,
            generic=generic,
        )
        for t in imported_types
    )


def is_iterable(value, include_strings=False):
    # type: (Any, bool) -> bool
    """
    Tell whether a value is an iterable or not.
    By default, strings are not considered iterables.

    :param value: Value.
    :param include_strings: Whether to consider strings as iterables.
    :return: True if iterable.
    """
    return isinstance(value, Iterable) and (  # type: ignore
        (not isinstance(value, six.string_types) or include_strings)
    )


def assert_is_instance(
    obj,  # type: _T
    types,  # type: Type[_T] | str | None | Iterable[Type[_T] | str | None]
    subtypes=True,  # type: bool
    extra_paths=(),  # type: Iterable[str]
    builtin_paths=None,  # type: Iterable[str] | None
    generic=True,  # type: bool
    typing=True,  # type: bool
):
    # type: (...) -> _T
    """
    Assert object is an instance of the provided types.

    :param obj: Object.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param extra_paths: Extra module paths in fallback order.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :param typing: Whether to check against typing.
    :raises ValueError: No types were provided.
    :raises TypeCheckError: Object is not an instance of provided types.
    """
    if not is_instance(
        obj,
        types,
        subtypes=subtypes,
        extra_paths=extra_paths,
        builtin_paths=builtin_paths,
        generic=generic,
        typing=typing,
    ):
        types = format_types(types)
        if not types:
            error = "no types were provided to perform assertion"
            raise ValueError(error)
        error = "got {!r} object, expected instance of {}{}".format(
            type(obj).__name__,
            ", ".join(repr(n) for n in type_names(types)),
            "" if subtypes else " (instances of subclasses are not accepted)",
        )
        raise TypeCheckError(error)
    return obj


def assert_is_subclass(
    cls,  # type: Type[_T]
    types,  # type: Type[_T] | str | None | Iterable[Type[_T] | str | None]
    subtypes=True,  # type: bool
    extra_paths=(),  # type: Iterable[str]
    builtin_paths=None,  # type: Iterable[str] | None
    generic=True,  # type: bool
    typing=True,  # type: bool
):
    # type: (...) -> Type[_T]
    """
    Assert a class is a subclass of the provided types.

    :param cls: Class.
    :param types: Types.
    :param subtypes: Whether to accept subtypes.
    :param extra_paths: Extra module paths in fallback order.
    :param builtin_paths: Builtin module paths in fallback order.
    :param generic: Whether to import generic.
    :param typing: Whether to check against typing.
    :raises ValueError: No types were provided.
    :raises TypeCheckError: Class is not a subclass of provided types.
    """
    if not isinstance(cls, type):
        types = format_types(types)
        error = "got instance of {!r}, expected {}{}{}".format(
            type(cls).__name__,
            "one of " if len(types) > 1 else "class ",
            ", ".join(repr(n) for n in type_names(types)),
            "" if subtypes else " (subclasses are not accepted)",
        )
        raise TypeCheckError(error)

    if not is_subclass(
        cls,
        types,
        subtypes=subtypes,
        extra_paths=extra_paths,
        builtin_paths=builtin_paths,
        generic=generic,
        typing=typing,
    ):
        types = format_types(types)
        if not types:
            error = "no types were provided to perform assertion"
            raise ValueError(error)

        error = "got {!r}, expected {}{}{}".format(
            cls.__name__,
            "one of " if len(types) > 1 else "class ",
            ", ".join(repr(n) for n in type_names(types)),
            "" if subtypes else " (subclasses are not accepted)",
        )
        raise TypeCheckError(error)

    return cast(Type[_T], cls)


def assert_is_callable(value):
    # type: (Callable[..., _T]) -> Callable[..., _T]
    """
    Assert a value is callable.

    :param value: Value.
    :raises TypeCheckError: Value is not a callable.
    """
    if not callable(value):
        error = "got non-callable {!r} object, expected a callable".format(type(value).__name__)
        raise TypeCheckError(error)
    return value


def assert_is_iterable(value, include_strings=False):
    # type: (Iterable[_T], bool) -> Iterable[_T]
    """
    Assert a value is iterable.
    By default, strings are not considered iterables.

    :param value: Value.
    :param include_strings: Whether to consider strings as iterables.
    :raises TypeCheckError: Value is not iterable.
    """
    if not is_iterable(value, include_strings=include_strings):
        error = "got non-iterable {!r} object, expected an iterable".format(type(value).__name__)
        raise TypeCheckError(error)
    return value



import copy
import pickle

import pytest  # noqa

from basicco.weak_reference import UniqueHashWeakReference, WeakReference


class Cls(object):
    pass


def test_empty_weak_reference():
    weak = WeakReference()
    assert weak() is None


def test_weak_reference():
    strong = Cls()
    strong_hash = hash(strong)
    weak_a = WeakReference(strong)
    weak_b = WeakReference(strong)
    assert weak_a() is strong
    assert weak_b() is strong
    assert hash(weak_a) == strong_hash
    assert hash(weak_b) == strong_hash
    assert weak_a == weak_b
    del strong
    assert weak_a() is None
    assert weak_b() is None
    assert hash(weak_a) == strong_hash
    assert hash(weak_b) == strong_hash
    assert weak_a == weak_b


def test_pickling():
    strong = Cls()
    weak = WeakReference(strong)
    unpickled_strong, unpickled_weak = pickle.loads(pickle.dumps((strong, weak)))
    assert isinstance(unpickled_strong, Cls)
    assert isinstance(unpickled_weak, WeakReference)
    assert unpickled_weak() is unpickled_strong
    assert type(pickle.loads(pickle.dumps(weak()))) is Cls


def test_caching():
    x = Cls()
    z = Cls()
    xr = WeakReference(x)
    zr = WeakReference(z)
    xr2 = WeakReference(x)
    zr2 = WeakReference(z)
    nr = WeakReference()
    nr2 = WeakReference()

    assert xr is xr2
    assert zr is zr2
    assert nr is nr2

    del z

    _x, _xr, _zr, _xr2, _zr2, _nr, _nr2 = pickle.loads(pickle.dumps((x, xr, zr, xr2, zr2, nr, nr2)))

    assert _xr is _xr2
    assert _zr is _zr2
    assert _nr is _nr2
    assert _nr is nr
    assert _nr2 is nr2
    assert _nr2 is nr
    assert _nr is nr2

    assert _xr() is _x
    assert _xr2() is _x


def test_deepcopy():
    x = Cls()
    z = Cls()
    xr = WeakReference(x)
    zr = WeakReference(z)
    xr2 = WeakReference(x)
    zr2 = WeakReference(z)
    nr = WeakReference()
    nr2 = WeakReference()

    assert xr is xr2
    assert zr is zr2
    assert nr is nr2

    del z

    _x, _xr, _zr, _xr2, _zr2, _nr, _nr2 = copy.deepcopy((x, xr, zr, xr2, zr2, nr, nr2))

    assert _xr is _xr2
    assert _zr is _zr2
    assert _nr is _nr2
    assert _nr is nr
    assert _nr2 is nr2
    assert _nr2 is nr
    assert _nr is nr2

    assert _xr() is _x
    assert _xr2() is _x


def test_hashing():
    class Unhashable(object):
        __hash__ = None

    unhashable = Unhashable()
    unhashable_ref = WeakReference(unhashable)
    hashable_ref = UniqueHashWeakReference(unhashable)

    assert unhashable_ref is not hashable_ref

    with pytest.raises(TypeError):
        hash(unhashable_ref)

    assert hash(hashable_ref) == object.__hash__(hashable_ref)

    assert WeakReference() is not UniqueHashWeakReference()


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa

from basicco.custom_repr import iterable_repr, mapping_repr


def test_custom_mapping_repr():
    mapping = {1: 4, "2": 3, 3: 2, "4": "1"}
    assert (
        mapping_repr(
            mapping,
            prefix="Mapping<",
            template="{key}=={value}",
            separator="; ",
            suffix=">",
            sorting=True,
            sort_key=lambda i: str(i[1]),
            reverse=True,
            key_repr=lambda k: repr("K" + str(k)),
            value_repr=lambda v: str("V" + str(v)),
        )
        == "Mapping<'K1'==V4; 'K2'==V3; 'K3'==V2; 'K4'==V1>"
    )

    items = ((1, 4), ("2", 3), (3, 2), ("4", "1"))
    assert (
        mapping_repr(
            items,
            prefix="Items:\n",
            template="    [{i}] {key: <3}: {value}",
            separator="\n",
            suffix="",
        )
        == """Items:
    [0] 1  : 4
    [1] '2': 3
    [2] 3  : 2
    [3] '4': '1'"""
    )


def test_custom_iterable_repr():
    iterable = ["a", 1, 2.0, "3.0", 4, None]
    assert iterable_repr(iterable) == repr(iterable)
    assert (
        iterable_repr(
            iterable,
            prefix="Iterable -",
            template=lambda i, value: "> {i}:{value} <".format(i=i, value=value),
            separator="-",
            suffix="-",
            sorting=True,
            sort_key=lambda v: str(v),
            reverse=True,
            value_repr=lambda v: str("V" + str(v)),
        )
        == "Iterable -> 0:Va <-> 1:VNone <-> 2:V4 <-> 3:V3.0 <-> 4:V2.0 <-> 5:V1 <-"
    )


if __name__ == "__main__":
    pytest.main()



import math

import pytest  # noqa

from basicco.data_class import DataClass, as_dict, as_tuple, deletes, field, fields, replace


class Vector(DataClass):
    x = field(required=False)  # type: float
    y = field()  # type: float
    name = field(eq=False, kw_only=True)  # type: str

    @property
    @field
    def mag(self):
        # type: () -> float
        return math.sqrt(self.x**2 + self.y**2)


class ImmutableVector(Vector):
    __kwargs__ = {"frozen": True}


def test_basic_data():
    vector_a = Vector(3, 4, name="vector_a")
    assert vector_a == Vector(3, 4, name="vector_b")

    assert repr(vector_a) == "Vector(3, 4, name='vector_a', mag=5.0)"

    vector_a.x = 4
    vector_a.y = 3
    assert vector_a != Vector(3, 4, "vector_a")

    assert "mag" in as_dict(vector_a)

    with pytest.raises(TypeError):
        hash(vector_a)

    assert list(fields(Vector).items()) == [
        ("x", Vector.__fields__["x"]),
        ("y", Vector.__fields__["y"]),
        ("name", Vector.__fields__["name"]),
        ("mag", Vector.__fields__["mag"]),
    ]


def test_immutable_basic_data():
    vector_m = Vector(3, 4, name="mut_vector")
    vector_m.x = 300
    vector_m.y = 400
    assert as_tuple(vector_m) == (300, 400, "mut_vector", 500)

    vector_a = ImmutableVector(3, 4, name="vector_a")
    assert vector_a == ImmutableVector(3, 4, name="vector_b")
    assert hash(vector_a) == hash(ImmutableVector(3, 4, name="vector_b"))

    assert replace(vector_a, x=30, y=40) == ImmutableVector(30, 40, name="vector_c")

    v = ImmutableVector(3, 4, name="vector_d")
    object.__delattr__(v, "x")
    assert deletes(vector_a, "x") == v

    assert list(fields(ImmutableVector).items()) == [
        ("x", ImmutableVector.__fields__["x"]),
        ("y", ImmutableVector.__fields__["y"]),
        ("name", ImmutableVector.__fields__["name"]),
        ("mag", ImmutableVector.__fields__["mag"]),
    ]


if __name__ == "__main__":
    pytest.main()



"""Utility to get mro bases."""

import collections
import inspect
import types

from tippo import GenericMeta, Type

__all__ = ["resolve_origin", "get_mro", "preview_mro"]


def resolve_origin(cls):
    # type: (Type | types.GenericAlias) -> Type
    """
    Resolve origin for generic alias.

    :param cls: Class or generic alias.
    :return: Original class.
    """
    if hasattr(cls, "__mro_entries__"):
        return cls.__mro_entries__(())[0]
    assert isinstance(cls, type)
    generic = cls
    while hasattr(generic, "__origin__"):
        generic = getattr(generic, "__origin__")
        if generic is not None:
            cls = generic
    return cls


def get_mro(cls):
    # type: (Type) -> tuple[Type, ...]
    """
    Get consistent mro amongst different python versions.
    This works even with generic classes.

    :param cls: Class.
    :return: Tuple of classes.
    """

    # Resolve generic class to its origin.
    cls = resolve_origin(cls)

    # Newer python versions.
    if GenericMeta is type:
        return tuple(inspect.getmro(cls))

    # Special logic to skip generic classes when GenericMeta is being used, for old python.
    mro = collections.deque()  # type: collections.deque[Type]
    for base in reversed(inspect.getmro(cls)):
        origin = resolve_origin(base)
        if origin in mro:
            continue
        mro.appendleft(base)
    return tuple(mro)


def _build_dummy_bases(bases, dummy_map, base_map):
    # type: (tuple[Type, ...], dict[Type, Type], dict[Type, Type]) -> tuple[Type, ...]
    dummy_bases = []
    for base in bases:

        # Reuse dummy base.
        if base in base_map:
            dummy_bases.append(base_map[base])
            continue

        # Make dummy base.
        if base is object or base is type:
            dummy_base = base
        else:
            dummy_base = type.__new__(
                type,
                base.__name__,
                _build_dummy_bases(base.__bases__, dummy_map, base_map),
                {},
            )

        # Remember and append dummy base.
        dummy_map[dummy_base] = base
        base_map[base] = dummy_base
        dummy_bases.append(dummy_base)

    return tuple(dummy_bases)


def preview_mro(*bases):
    # type: (*Type) -> tuple[Type, ...]
    """
    Preview the mro before building the actual class.

    :param bases: Bases.
    :return: Tuple of classes.
    """
    dummy_map = {}  # type: dict[Type, Type]
    base_map = {}  # type: dict[Type, Type]
    dummy_bases = _build_dummy_bases(bases, dummy_map, base_map)
    dummy_cls = type.__new__(type, "Dummy", dummy_bases, {})
    return tuple(dummy_map[b] for b in get_mro(dummy_cls)[1:])



import pytest  # noqa

from basicco.safe_repr import safe_repr


def test_safe_repr():
    class Class(object):
        @safe_repr
        def __repr__(self):
            raise RuntimeError("oh oh")

    obj = Class()
    standard_repr = object.__repr__(obj)
    assert repr(obj) == standard_repr[:-1] + "; repr failed due to 'RuntimeError: oh oh'>"


if __name__ == "__main__":
    pytest.main()



"""Python 2.7 compatible implementation of slotted data classes based on a metaclass."""

import copy
from collections import OrderedDict
from enum import Enum

import six
from tippo import Any, Callable, Final, Literal, Type, TypedDict, TypeVar, cast, dataclass_transform

from ._bases import SlottedBase, SlottedBaseMeta
from .custom_repr import mapping_repr
from .dynamic_code import generate_unique_filename, make_function
from .get_mro import preview_mro
from .mangling import mangle, unmangle
from .mapping_proxy import MappingProxyType
from .recursive_repr import recursive_repr
from .runtime_final import final
from .safe_repr import safe_repr

__all__ = ["field", "fields", "replace", "deletes", "as_items", "as_dict", "as_tuple", "DataClassMeta", "DataClass"]


T = TypeVar("T")


class _MissingType(Enum):
    MISSING = "MISSING"


class _DefaultType(Enum):
    DEFAULT = "DEFAULT"


# Sentinel value for missing values.
_MISSING = _MissingType.MISSING

# Sentinel value for default values.
_DEFAULT = _DefaultType.DEFAULT


# Global field count for ordering.
_field_count = 0


@final
class _Field(SlottedBase):
    """Field type."""

    __slots__ = (
        "id",
        "property",
        "default",
        "default_factory",
        "converter",
        "init",
        "kw_only",
        "repr",
        "eq",
        "hash",
        "frozen",
        "order",
        "required",
    )

    def __init__(
        self,
        default,  # type: T | _MissingType
        default_factory,  # type: Callable[..., T] | _MissingType
        converter,  # type: Callable[[Any], T] | None
        init,  # type: bool
        kw_only,  # type: bool
        repr,  # type: bool
        eq,  # type: bool
        hash,  # type: bool | None
        frozen,  # type: bool
        order,  # type: bool | None
        required,  # type: bool | None
    ):
        # type: (...) -> None

        # Ensure single default source.
        if default is not _MISSING and default_factory is not _MISSING:
            error = "can't declare both 'default' and 'default_factory'"
            raise ValueError(error)

        # Ensure safe hash.
        eq = bool(eq)
        if hash is None:
            hash = eq
        hash = bool(hash)
        if hash and not eq:
            error = "can't contribute to the hash when not contributing to the eq"
            raise ValueError(error)

        # Ensure safe required and order.
        if order is None:
            if required is None:
                required = True
            order = eq and required
        else:
            order = bool(order)
        if order and required is None:
            required = True
        else:
            required = bool(required)
        if order and not eq:
            error = "can't contribute to order if it's not contributing to the eq"
            raise ValueError(error)
        if order and not required:
            error = "can't contribute to order if it's not required"
            raise ValueError(error)

        # Cast parameters.
        init = bool(init)
        kw_only = bool(kw_only)
        repr = bool(repr)
        frozen = bool(frozen)

        # Increment field count.
        global _field_count
        _field_count += 1

        self.default = default  # type: T | _MissingType
        self.default_factory = default_factory  # type: Callable[..., T] | _MissingType
        self.converter = converter  # type: Callable[[Any], T] | None
        self.init = init  # type: bool
        self.kw_only = kw_only  # type: bool
        self.repr = repr  # type: bool
        self.eq = eq  # type: bool
        self.hash = hash  # type: bool
        self.frozen = frozen  # type: bool
        self.order = order  # type: bool
        self.required = required  # type: bool
        self.id = _field_count  # type: int
        self.property = None  # type: property | None

    def __call__(self, fget):
        # type: (T) -> T
        """
        Allows field to be used as a decorator for a property's getter function.

        :param fget: Getter method function.
        :return: Decorated getter method function.
        """
        object.__setattr__(fget, "__field__", self)
        return fget

    def __hash__(self):
        return hash(tuple(self.__items()))

    def __eq__(self, other):
        return type(self) is type(other) and self.__items() == other.__items()

    def __setattr__(self, name, value):
        if hasattr(self, name) and not name.startswith("_"):
            error = "can't set read-only attribute {!r}".format(name)
            raise AttributeError(error)
        super(_Field, self).__setattr__(name, value)

    def __delattr__(self, name):
        if hasattr(self, name) and not name.startswith("_"):
            error = "can't delete read-only attribute {!r}".format(name)
            raise AttributeError(error)
        super(_Field, self).__delattr__(name)

    @safe_repr
    @recursive_repr
    def __repr__(self):
        # type: () -> str
        return mapping_repr(
            self.__items(),
            template="{key}={value}",
            prefix="Field(",
            suffix=")",
            key_repr=str,
            value_repr=repr,
        )

    def __items(self):
        return [
            ("default", self.default),
            ("default_factory", self.default_factory),
            ("converter", self.converter),
            ("init", self.init),
            ("kw_only", self.kw_only),
            ("repr", self.repr),
            ("eq", self.eq),
            ("hash", self.hash),
            ("frozen", self.frozen),
            ("order", self.order),
            ("id", self.id),
            ("property", self.property),
        ]


def field(
    default=_MISSING,  # type: T | _MissingType
    default_factory=_MISSING,  # type: Callable[..., T] | _MissingType
    converter=None,  # type: Callable[[Any], T] | None
    init=True,  # type: bool
    kw_only=False,  # type: bool
    repr=True,  # type: bool
    eq=True,  # type: bool
    hash=None,  # type: bool | None
    frozen=False,  # type: bool
    order=None,  # type: bool | None
    required=None,  # type: bool | None
):
    # type: (...) -> T
    """
    Define a field.

    :param default: Default value.
    :param default_factory: Default value factory.
    :param converter: Value converter.
    :param init: Whether to include in `__init__`.
    :param kw_only: Whether to only accept keyword argument initialization.
    :param repr: Whether to include in `__repr__`.
    :param eq: Whether to include in `__eq__`.
    :param hash: Whether to include in `__hash__`.
    :param frozen: Whether should be frozen (immutable).
    :param order: Whether to include in order methods.
    :param required: Whether to require a value after initialization/replacement.
    """
    _field = _Field(
        default=default,
        default_factory=default_factory,
        converter=converter,
        init=init,
        kw_only=kw_only,
        repr=repr,
        eq=eq,
        hash=hash,
        frozen=frozen,
        order=order,
        required=required,
    )
    return cast(T, _field)


def fields(cls):
    # type: (Type[DataClass]) -> OrderedDict[str, _Field]
    """
    Get data class fields.

    :param cls: Data class.
    :return: Ordered dictionary of fields.
    """
    return OrderedDict(cls.__fields__)


def replace(__data, **replacements):
    # type: (DataClass, **Any) -> DataClass
    """
    Make a copy with replaced field values.

    :param __data: Data.
    :param replacements: Field value replacements.
    :return: Data with replacements.
    """
    new_data = copy.copy(__data)
    for name, value in replacements.items():
        converter = type(__data).__fields__[name].converter
        if converter is not None:
            value = converter(value)
        object.__setattr__(new_data, name, value)
    return new_data


def deletes(__data, *deletions):
    # type: (DataClass, *str) -> DataClass
    """
    Make a copy with deleted field values.

    :param __data: Data.
    :param deletions: Fields to delete.
    :return: Data with deletions.
    """
    cls = type(__data)
    required = set(n for n, f in cls.__fields__.items() if f.required).intersection(deletions)
    if required:
        error = "can't delete required field(s) {}".format(", ".join(repr(r) for r in required))
        raise AttributeError(error)
    new_data = copy.copy(__data)
    for name in deletions:
        object.__delattr__(new_data, name)
    return new_data


def _as_items(converter, data, recursively):
    # type: (Callable[..., Any], DataClass, bool) -> list[tuple[str, Any]]
    items = []
    for field_name, field_ in type(data).__fields__.items():
        value = getattr(data, field_name, _MISSING)
        if value is _MISSING:
            continue
        if recursively:
            if isinstance(value, DataClass):
                value = converter(value, True)
            elif isinstance(value, (tuple, list, set, frozenset)):
                value = type(value)(converter(v, True) if isinstance(v, DataClass) else v for v in value)
            elif isinstance(value, (dict, MappingProxyType)):
                value = type(value)(
                    dict((k, converter(v, True)) if isinstance(v, DataClass) else (k, v) for k, v in value.items())
                )
        items.append((field_name, value))
    return items


def as_items(data, recursively=True):
    # type: (DataClass, bool) -> list[tuple[str, Any]]
    """
    Convert data to ordered items.

    :param data: Data.
    :param recursively: Whether to convert recursively.
    :return: Items.
    """
    return _as_items(as_items, data, recursively)


def as_dict(data, recursively=True):
    # type: (DataClass, bool) -> dict[str, Any]
    """
    Convert data to dictionary.

    :param data: Data.
    :param recursively: Whether to convert recursively.
    :return: Dictionary.
    """
    return dict(_as_items(as_dict, data, recursively))


def as_tuple(data, recursively=True):
    # type: (DataClass, bool) -> tuple[Any, ...]
    """
    Convert data to a tuple.

    :param data: Data.
    :param recursively: Whether to convert recursively.
    :return: Tuple.
    """
    return tuple(list(zip(*_as_items(as_tuple, data, recursively)))[1])


def _has_default(field_):
    # type: (_Field) -> bool
    """
    Get whether field has default value.

    :param field_: Field.
    :return: True if has default value.
    """
    return field_.default is not _MISSING


def _has_default_factory(field_):
    # type: (_Field) -> bool
    """
    Get whether field has default default_factory.

    :param field_: Field.
    :return: True if has default default_factory.
    """
    return field_.default_factory is not _MISSING


def _has_any_default(field_):
    # type: (_Field) -> bool
    """
    Get whether field has default value (or default_factory).

    :param field_: Field.
    :return: True if has default.
    """
    return _has_default(field_) or _has_default_factory(field_)


def _make_init(module, cls_fields, cls_name, kw_only):
    # type: (str, OrderedDict[str, _Field], str, bool) -> Callable[..., None]
    """
    Make `__init__` method function.

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param kw_only: Whether to accept keyword arguments only.
    :param cls_name: Class name.
    :return: Method function.
    """
    script = "def __init__(self"
    globs = {"object": object}  # type: dict[str, Any]
    params = []  # type: list[str]
    lines = []  # type: list[str]
    required_lines = []  # type: list[str]
    seen_default = None  # type: str | None
    for field_name, field_ in cls_fields.items():

        # Add required field check.
        if field_.required and not field_.init:
            required_lines.append("if not hasattr(self, '{f}'):".format(f=field_name))
            required_lines.append("    error = 'missing value for required attribute {f}'".format(f=field_name))
            required_lines.append("    raise TypeError(error)")

        # Skip non-init fields.
        if not field_.init:
            continue

        # Converter variable name.
        converter_var = "{f}_converter".format(f=field_name)
        while converter_var in cls_fields:
            converter_var = "{}_".format(converter_var)
        if field_.converter is not None:
            globs[converter_var] = field_.converter

        # Non-default field.
        if not _has_any_default(field_):

            # Not forcing keyword arguments.
            if not kw_only and not field_.kw_only:

                # Non-default field after default field, error out.
                if seen_default is not None:
                    error = "non-default field {!r} declared after default field {!r}".format(field_name, seen_default)
                    raise TypeError(error)

                # Simple argument.
                params.append(field_name)
                if field_.converter is None:
                    lines.append("object.__setattr__(self, '{f}', {f})".format(f=field_name))
                else:
                    lines.append("object.__setattr__(self, '{f}', {c}({f}))".format(c=converter_var, f=field_name))

            # Forcing keyword arguments.
            else:
                seen_default = field_name
                globs["_MISSING"] = _MISSING
                params.append("{f}=_MISSING".format(f=field_name))
                lines.append("if {f} is _MISSING:".format(f=field_name))
                lines.append("    error = 'missing value for required keyword argument {f}'".format(f=field_name))
                lines.append("    raise TypeError(error)")
                if field_.converter is None:
                    lines.append("object.__setattr__(self, '{f}', {f})".format(f=field_name))
                else:
                    lines.append("object.__setattr__(self, '{f}', {c}({f}))".format(c=converter_var, f=field_name))

        # Default field.
        else:
            seen_default = field_name

            # Default variable name.
            default_var = "DEFAULT_{f}".format(f=field_name.upper())
            while default_var in cls_fields:
                default_var = "{}_".format(default_var)

            # Field has a default value.
            if _has_default(field_):
                globs[default_var] = field_.default
                params.append("{}={}".format(field_name, default_var))
                if field_.converter is None:
                    lines.append("object.__setattr__(self, '{f}', {f})".format(f=field_name))
                else:
                    lines.append("object.__setattr__(self, '{f}', {c}({f}))".format(c=converter_var, f=field_name))

            # Field has a default factory.
            else:
                assert _has_default_factory(field_)
                globs[default_var] = _DEFAULT
                params.append("{}={}".format(field_name, default_var))
                lines.append("if {} is {}:".format(field_name, default_var))
                if field_.converter is None or field_.converter is field_.default_factory:
                    lines.append(
                        "    object.__setattr__(self, '{f}', type(self).__fields__['{f}'].default_factory())".format(
                            f=field_name
                        )
                    )
                else:
                    lines.append(
                        "    object.__setattr__(self, '{f}', "
                        "{c}(type(self).__fields__['{f}'].default_factory()))".format(c=converter_var, f=field_name)
                    )
                lines.append("else:")
                if field_.converter is None:
                    lines.append("    object.__setattr__(self, '{f}', {f})".format(f=field_name))
                else:
                    lines.append("    object.__setattr__(self, '{f}', {c}({f}))".format(c=converter_var, f=field_name))

    script += (", " + ", ".join(params) if params else "") + "):\n"
    script += ("    " + "\n    ".join(lines) + "\n") if lines else ""
    script += "    if hasattr(self, '__post_init__'):\n"
    script += "        self.__post_init__()\n"
    script += ("    " + "\n    ".join(required_lines) + "\n") if lines else ""

    return make_function(
        "__init__",
        script,
        globs,
        generate_unique_filename("__init__", module, cls_name),
        module,
    )


def _make_eq(module, cls_fields, cls_name):
    # type: (str, OrderedDict[str, _Field], str) -> Callable[..., None]
    """
    Make `__eq__` method function.

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param cls_name: Class name.
    :return: Method function.
    """
    script = "def __eq__(self, other):\n"
    script += "    if self is other:\n"
    script += "        return True\n"
    script += "    if type(self) is not type(other):\n"
    script += "        return False"
    globs = {"_MISSING": _MISSING}  # type: dict[str, Any]
    lines = []  # type: list[str]
    for field_name, field_ in cls_fields.items():
        if not field_.eq:
            continue
        assert field_.hash
        lines.append("if getattr(self, '{f}', _MISSING) != getattr(other, '{f}', _MISSING):".format(f=field_name))
        lines.append("    return False")

    script += ("\n    " + "\n    ".join(lines)) if lines else ""
    script += "\n    return True\n"

    return make_function(
        "__eq__",
        script,
        globs,
        generate_unique_filename("__eq__", module, cls_name),
        module,
    )


def _make_repr(module, cls_fields, cls_name, kw_only):  # noqa
    # type: (str, OrderedDict[str, _Field], str, bool) -> Callable[..., None]
    """
    Make `__repr__` method function.

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param cls_name: Class name.
    :param kw_only: Whether to accept keyword arguments only.
    :return: Method function.
    """
    script = "@safe_repr\n"
    script += "@recursive_repr\n"
    script += "def __repr__(self):\n"
    script += "    repr_str = '{}('\n".format(cls_name)
    script += "    parts = []\n"
    for field_name, field_ in cls_fields.items():
        script += "    if hasattr(self, '{f}'):\n".format(f=field_name)
        if kw_only or _has_any_default(field_) or field_.kw_only:
            script += "        parts.append('" + field_name + "={!r}'.format(getattr(self, '" + field_name + "')))\n"
        else:
            script += "        parts.append('{!r}'.format(getattr(self, '" + field_name + "')))\n"
    script += "    repr_str += ', '.join(parts) + ')'\n"
    script += "    return repr_str\n"
    globs = {"safe_repr": safe_repr, "recursive_repr": recursive_repr, "type": type}  # type: dict[str, Any]

    return make_function(
        "__repr__",
        script,
        globs,
        generate_unique_filename("__repr__", module, cls_name),
        module,
    )


def _make_hash(module, cls_fields, cls_name, frozen_cls):
    # type: (str, OrderedDict[str, _Field], str, bool) -> Callable[..., None] | None
    """
    Make `__hash__` method function (or None for mutable classes).

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param cls_name: Class name.
    :param frozen_cls: Whether class is frozen.
    :return: Method function or None.
    """
    script = "def __hash__(self):\n"
    script += "    return hash(("
    globs = {"_MISSING": _MISSING}  # type: dict[str, Any]
    lines = []  # type: list[str]
    for field_name, field_ in cls_fields.items():
        if not field_.hash:
            continue
        if not frozen_cls and not field_.frozen:
            return None
        lines.append("getattr(self, '{f}', _MISSING)".format(f=field_name))

    script += ("\n        " + ",\n        ".join(lines) + "\n") if lines else ""
    script += "    ))\n"

    return make_function(
        "__hash__",
        script,
        globs,
        generate_unique_filename("__eq__", module, cls_name),
        module,
    )


def _make_setattr(module, cls_fields, cls_name, frozen_cls):
    # type: (str, OrderedDict[str, _Field], str, bool) -> Callable[..., None]
    """
    Make `__setattr__` method function.

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param cls_name: Class name.
    :param frozen_cls: Whether class is frozen.
    :return: Method function.
    """
    script = "def __setattr__(self, name, value):\n"
    globs = {"object": object, "type": type}  # type: dict[str, Any]
    lines = []  # type: list[str]
    if frozen_cls:
        lines.append("error = 'class {} is frozen'".format(cls_name))
        lines.append("raise AttributeError(error)")
    else:
        if any(f.frozen for f in cls_fields.values()):
            lines.append("if name in {!r}:".format(set(n for n, f in cls_fields.items() if f.frozen)))
            lines.append("    error = 'field {!r} is frozen'.format(name)")
            lines.append("    raise AttributeError(error)")
        if any(f.converter is not None for f in cls_fields.values()):
            lines.append("if name in {!r}:".format(set(n for n, f in cls_fields.items() if f.converter is not None)))
            lines.append("    value = type(self).__fields__[name].converter(value)")
        lines.append("object.__setattr__(self, name, value)")

    script += (("    " + "\n    ".join(lines)) if lines else "") + "\n"

    return make_function(
        "__setattr__",
        script,
        globs,
        generate_unique_filename("__setattr__", module, cls_name),
        module,
    )


def _make_delattr(module, cls_fields, cls_name, frozen_cls):
    # type: (str, OrderedDict[str, _Field], str, bool) -> Callable[..., None]
    """
    Make `__delattr__` method function.

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param cls_name: Class name.
    :param frozen_cls: Whether class is frozen.
    :return: Method function.
    """
    script = "def __delattr__(self, name):\n"
    globs = {"object": object}  # type: dict[str, Any]
    lines = []  # type: list[str]
    if frozen_cls:
        lines.append("error = 'class {} is frozen'".format(cls_name))
        lines.append("raise AttributeError(error)")
    else:
        lines.append("if name in {!r}:".format(set(n for n, f in cls_fields.items() if f.frozen)))
        lines.append("    error = 'field {!r} is frozen'.format(name)")
        lines.append("    raise AttributeError(error)")
        lines.append("object.__delattr__(self, name)")

    script += (("    " + "\n    ".join(lines)) if lines else "") + "\n"

    return make_function(
        "__delattr__",
        script,
        globs,
        generate_unique_filename("__delattr__", module, cls_name),
        module,
    )


_order_operators = {
    "__lt__": "<",
    "__le__": "<=",
    "__gt__": ">",
    "__ge__": ">=",
}


def _make_order(
    module,  # type: str
    cls_fields,  # type: OrderedDict[str, _Field]
    cls_name,  # type: str
    method_name,  # type: Literal["__lt__", "__le__", "__gt__", "__ge__"]
):
    # type: (...) -> Callable[..., None]
    """
    Make order method function.

    :param module: Module.
    :param cls_fields: Ordered dictionary of named cls_fields.
    :param cls_name: Class name.
    :param method_name: Method name.
    :return: Method function.
    """
    script = "def {}(self, other):\n".format(method_name)
    script += "    cls = type(self)\n"
    script += "    if cls is not type(other):\n"
    script += "        return NotImplemented\n"
    globs = {"object": object}  # type: dict[str, Any]
    self_parts = []  # type: list[str]
    other_parts = []  # type: list[str]
    for field_name, field_ in cls_fields.items():
        if not field_.order:
            continue

        self_parts.append("self.{f}".format(f=field_name))
        other_parts.append("other.{f}".format(f=field_name))

    script += "    self_values = (" + ", ".join(self_parts) + ")\n"
    script += "    other_values = (" + ", ".join(other_parts) + ")\n"
    script += "    return self_values {} other_values\n".format(_order_operators[method_name])

    return make_function(
        method_name,
        script,
        globs,
        generate_unique_filename(method_name, module, cls_name),
        module,
    )


def _make_match_args(cls_fields):
    # type: (OrderedDict[str, _Field]) -> tuple[str, ...]
    """
    Make `__match_args__` tuple class variable.

    :param cls_fields: Ordered dictionary of named cls_fields.
    :return: Match args tuple.
    """
    match_args = []
    for field_name, field_ in cls_fields.items():
        if not field_.init:
            continue
        match_args.append(field_name)
    return tuple(match_args)


DataClassParams = TypedDict(
    "DataClassParams",
    {
        "kw_only": bool,
        "init": bool,
        "repr": bool,
        "eq": bool,
        "hash": bool,
        "frozen": bool,
        "unsafe_hash": bool,
        "order": bool,
        "match_args": bool,
    },
    total=False,
)


@dataclass_transform(field_specifiers=(field,))
class DataClassMeta(SlottedBaseMeta):
    """Metaclass for :class:`DataClass`."""

    @staticmethod
    def __new__(mcs, name, bases, dct, **kwargs):  # noqa
        # type: (...) -> DataClassMeta
        dct = dict(dct)

        # Gather fields for this class.
        __fields = dct[mangle("__fields", name)] = OrderedDict()
        for member_name, member in list(dct.items()):

            # Regular field.
            if isinstance(member, _Field):
                __fields[member_name] = member

            # Property field.
            elif isinstance(member, property) and member.fget is not None:
                if hasattr(member.fget, "__field__"):
                    field_ = member.fget.__field__
                    object.__delattr__(member.fget, "__field__")
                elif isinstance(member.fget, _Field):
                    field_ = member.fget
                    assert field_.default is not _MISSING  # noqa
                    dct[member_name] = member.getter(field_.default)  # noqa
                else:
                    continue
                object.__setattr__(field_, "property", member)
                object.__setattr__(field_, "init", member.fset is not None and field_.init)
                object.__setattr__(field_, "frozen", (member.fset is None and member.fdel is None) or field_.frozen)
                __fields[member_name] = field_

        # Convert non-property fields to slots.
        dct["__slots__"] = tuple(
            sorted(
                set(dct.get("__slots__", ())).union(
                    unmangle(n, name) for n, f in __fields.items() if f.property is None and dct.pop(n)
                )
            )
        )

        # Loop through mro.
        mro = [
            (b.__name__, b.__dict__, b.__dict__.get(mangle("__fields", b.__name__), {}), isinstance(b, DataClassMeta))
            for b in preview_mro(*bases)
        ] + [(name, dct, __fields, True)]
        all_fields = {}
        field_ids = {}
        for base_name, base_dct, base_fields, is_data in mro:

            # Prevent bad overrides.
            for member_name, member in base_dct.items():
                if isinstance(member, _Field):
                    assert not is_data
                    error = "non-data class {!r} defines a {!r} field".format(base_name, member_name)
                    raise TypeError(error)
                if member_name not in base_fields and member_name in all_fields:
                    error = "class {!r} overrides field {!r} with non-field {!r} object".format(
                        base_name, member_name, type(member).__name__
                    )
                    raise TypeError(error)

            # Keep track of fields.
            for field_name, field_ in base_fields.items():
                if field_name in ("object", "_MISSING", "_DEFAULT"):
                    error = "invalid field name {!r}".format(field_name)
                    raise ValueError(error)
                all_fields[field_name] = field_
                if field_name not in field_ids:
                    field_ids[field_name] = field_.id

        # Order fields by id and store them in a ordered dict.
        dct["__fields__"] = OrderedDict(sorted(all_fields.items(), key=lambda i: field_ids[i[0]]))

        # Build class and return it.
        try:
            return super(DataClassMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)
        except TypeError as e:
            exc = TypeError(e)
            six.raise_from(exc, None)
            raise exc


class DataClass(six.with_metaclass(DataClassMeta, SlottedBase)):
    """Basic data class."""

    __slots__ = ()
    __fields__ = OrderedDict()  # type: Final[OrderedDict[str, _Field]]
    __params__ = {}  # type: Final[DataClassParams]
    __kwargs__ = {}  # type: DataClassParams

    def __init_subclass__(
        cls,
        kw_only=None,  # type: bool | None
        init=None,  # type: bool | None
        repr=None,  # type: bool | None
        eq=None,  # type: bool | None
        hash=None,  # type: bool | None
        frozen=None,  # type: bool | None
        unsafe_hash=False,  # type: bool
        order=None,  # type: bool | None
        match_args=None,  # type: bool | None
    ):
        # type: (...) -> None
        """
        Initialize subclass with parameters.

        :param kw_only: Force keyword-only arguments in generated `__init__` method.
        :param init: Whether to generate `__init__` method.
        :param repr: Whether to generate `__repr__` method.
        :param eq: Whether to generate `__eq__` method.
        :param hash: Whether to generate `__hash__` method.
        :param frozen: Whether class should be immutable.
        :param unsafe_hash: Whether to allow manual declaration of potentially unsafe `__hash__` method.
        :param order: Whether to generate order methods.
        :param match_args: Whether to generate `__match_args__`.
        """

        # Merge parameters.
        type.__setattr__(
            cls,
            "__params__",
            dict(
                cls.__params__,
                **dict(
                    (n, bool(v))
                    for n, v in {
                        "kw_only": kw_only,
                        "init": init,
                        "repr": repr,
                        "eq": eq,
                        "hash": hash,
                        "frozen": frozen,
                        "order": order,
                        "match_args": match_args,
                    }.items()
                    if v is not None
                )
            ),
        )

        # __init__.
        if cls.__params__.get("init", True):
            if "__init__" in cls.__dict__:
                error = "auto generated '__init__' method can't be overriden without disabling 'init' parameter"
                raise TypeError(error)
            type.__setattr__(
                cls,
                "__init__",
                _make_init(cls.__module__, cls.__fields__, cls.__name__, cls.__params__.get("kw_only", False)),
            )
        elif "__init__" not in cls.__dict__:
            type.__setattr__(cls, "__init__", object.__init__)

        # __repr__.
        if cls.__params__.get("repr", True):
            if "__repr__" in cls.__dict__:
                error = "auto generated '__repr__' method can't be overriden without disabling 'repr' parameter"
                raise TypeError(error)
            type.__setattr__(
                cls,
                "__repr__",
                _make_repr(cls.__module__, cls.__fields__, cls.__name__, cls.__params__.get("kw_only", False)),
            )
        elif "__repr__" not in cls.__dict__:
            type.__setattr__(cls, "__repr__", object.__init__)

        # __eq__.
        if cls.__params__.get("eq", True):
            if "__eq__" in cls.__dict__:
                error = "auto generated '__eq__' method can't be overriden without disabling 'eq' parameter"
                raise TypeError(error)
            type.__setattr__(cls, "__eq__", _make_eq(cls.__module__, cls.__fields__, cls.__name__))
        elif "__eq__" not in cls.__dict__ and "__hash__" not in cls.__dict__:
            type.__setattr__(cls, "__eq__", object.__eq__)

        # __hash__.
        frozen_cls = cls.__params__.get("frozen", False)
        if cls.__params__.get("hash", True):
            if "__hash__" in cls.__dict__:
                error = (
                    "auto generated '__hash__' method can't be overriden without disabling parameters 'hash' & 'eq' "
                    "and enabling 'unsafe_hash'"
                )
                raise TypeError(error)
            type.__setattr__(cls, "__hash__", _make_hash(cls.__module__, cls.__fields__, cls.__name__, frozen_cls))
        elif "__hash__" not in cls.__dict__:
            if cls.__eq__ is object.__eq__:
                type.__setattr__(cls, "__hash__", object.__hash__)
            else:
                type.__setattr__(cls, "__hash__", None)
        elif not unsafe_hash:
            error = "it's unsafe to declare your own '__hash__' method"
            raise TypeError(error)
        elif cls.__params__.get("eq", True):
            error = "can't declare your own '__hash__' method without disabling the 'eq' parameter"
            raise TypeError(error)
        elif "__eq__" not in cls.__dict__:
            error = "when declaring your own '__hash__' method you also need to declare '__eq__'"
            raise TypeError(error)

        # __setattr__ & __delattr__.
        for method, method_maker in (("__setattr__", _make_setattr), ("__delattr__", _make_delattr)):
            if frozen_cls or any(
                (f.frozen or f.converter is not None)
                for f in cls.__fields__.values()
                if (f.property is None or f.converter is not None)
            ):
                if method in cls.__dict__:
                    error = (
                        "auto generated '{}' method can't be overriden without disabling 'frozen' parameter "
                        "for class and all of its fields".format(method)
                    )
                    raise TypeError(error)
                type.__setattr__(cls, method, method_maker(cls.__module__, cls.__fields__, cls.__name__, frozen_cls))
            elif method not in cls.__dict__:
                type.__setattr__(cls, method, getattr(object, method))

        # order.
        for method_name, order_operator in _order_operators.items():
            if cls.__params__.get("order", False):
                if method_name in cls.__dict__:
                    error = "auto generated {!r} method can't be overriden without disabling 'order' parameter".format(
                        method_name
                    )
                    raise TypeError(error)
                type.__setattr__(
                    cls,
                    method_name,
                    _make_order(cls.__module__, cls.__fields__, cls.__name__, method_name),  # type: ignore
                )
            elif method_name not in cls.__dict__:
                type.__setattr__(cls, method_name, getattr(object, method_name))

        # __match_args__.
        if cls.__params__.get("match_args", True):
            if "__match_args__" in cls.__dict__:
                error = (
                    "auto generated '__match_args__' tuple can't be overriden without disabling 'match_args' parameter"
                )
                raise TypeError(error)
            type.__setattr__(cls, "__match_args__", _make_match_args(cls.__fields__))
        elif "__match_args__" not in cls.__dict__:
            type.__setattr__(cls, "__match_args__", ())

    def __init__(self, *args, **kwargs):
        pass

    def __post_init__(self):
        pass



"""Mapping Proxy type (read-only) for older Python versions."""

__all__ = ["MappingProxyType"]


try:
    from types import MappingProxyType

except ImportError:
    from tippo import Iterator, Mapping, TypeVar  # type: ignore

    KT = TypeVar("KT")
    VT = TypeVar("VT")

    class _MappingProxyType(Mapping[KT, VT]):  # type: ignore
        __slots__ = ("__mapping",)

        def __init__(self, mapping):
            # type: (Mapping[KT, VT]) -> None
            self.__mapping = mapping

        def __getitem__(self, item):
            # type: (KT) -> VT
            return self.__mapping[item]

        def __len__(self):
            # type: () -> int
            return len(self.__mapping)

        def __iter__(self):
            # type: () -> Iterator[KT]
            for key in self.__mapping:
                yield key

    type.__setattr__(_MappingProxyType, "__name__", "MappingProxyType")
    type.__setattr__(_MappingProxyType, "__qualname__", "MappingProxyType")
    globals()["MappingProxyType"] = _MappingProxyType



"""Named tuple utilities."""

import functools

__all__ = ["defaults"]


_MISSING = object()


def _check_missing_decorator(new_method):
    @functools.wraps(new_method)
    def decorated(*args, **kwargs):
        self = new_method(*args, **kwargs)
        missing_fields = [f for f in type(self)._fields if getattr(self, f) is _MISSING]  # noqa
        for field in type(self)._fields:  # noqa
            if getattr(self, field) is _MISSING:
                error = "{}.__new__() missing {} required positional argument{}: {}".format(
                    type(self).__name__,
                    len(missing_fields),
                    "s" if len(missing_fields) > 1 else "",
                    ", ".join(repr(f) for f in missing_fields),
                )
                raise TypeError(error)
        return self

    return decorated


def defaults(**values):
    """
    Sets default values for named tuple classes.

    :param values: Default values.
    :return: Decorator that adds default values to a named tuple class.
    """

    def decorator(cls):
        if values:
            kw = None
            for field in cls._fields:
                if field in values:
                    kw = field
                elif kw is not None:
                    error = "positional field {!r} after default field {!r}".format(field, kw)
                    raise TypeError(error)
            cls.__new__.__defaults__ = (_MISSING,) * len(cls._fields)
            prototype = cls(**values)
            cls.__new__.__defaults__ = tuple(prototype)
            type.__setattr__(cls, "__new__", staticmethod(_check_missing_decorator(cls.__new__)))
        return cls

    return decorator



"""
Backport of `contextvars` for Python 2.7.
Based on `MagicStack/contextvars <https://github.com/MagicStack/contextvars>`_.
"""

__all__ = ["ContextVar", "Context", "Token", "copy_context"]


try:
    from contextvars import Context, ContextVar, Token, copy_context  # noqa

except ImportError:
    import enum
    import threading

    import pyrsistent
    import six
    from pyrsistent.typing import PMap
    from tippo import Any, Callable, Generic, GenericMeta, Iterator, Mapping, TypeVar, overload

    class _NoDefaultType(enum.Enum):
        NO_DEFAULT = "NO_DEFAULT"

    _NO_DEFAULT = _NoDefaultType.NO_DEFAULT
    _MODULE = __name__

    T = TypeVar("T")
    RT = TypeVar("RT")

    class _ContextVarMeta(GenericMeta):
        def __new__(mcs, names, bases, dct, **kwargs):
            cls = super(_ContextVarMeta, mcs).__new__(mcs, names, bases, dct, **kwargs)
            if cls.__module__ != _MODULE or cls.__name__ != "_ContextVar":
                raise TypeError("type '_ContextVar' is not an acceptable base type")
            return cls

    class _ContextVar(six.with_metaclass(_ContextVarMeta, Generic[T])):
        __slots__ = ("__weakref__", "_name", "_default")

        def __init__(self, name, default=_NO_DEFAULT):
            # type: (str, T | _NoDefaultType) -> None
            if not isinstance(name, str):
                raise TypeError("context variable name must be a str")
            self._name = name
            self._default = default

        @property
        def name(self):
            # type: () -> str
            return self._name

        @overload
        def get(self, default=_NO_DEFAULT):
            # type: (T | _NoDefaultType) -> T
            pass

        @overload
        def get(self, default=_NO_DEFAULT):
            # type: (Any) -> Any
            pass

        def get(self, default=_NO_DEFAULT):
            ctx = _get_context()
            try:
                return ctx[self]
            except KeyError:
                pass

            if default is not _NO_DEFAULT:
                return default

            if self._default is not _NO_DEFAULT:
                return self._default

            raise LookupError()

        def set(self, value):
            # type: (T) -> _Token[T]
            ctx = _get_context()
            data = ctx._data  # noqa
            try:
                old_value = data[self]
            except KeyError:
                old_value = _Token.MISSING

            updated_data = data.set(self, value)
            ctx._data = updated_data
            return _Token(ctx, self, old_value)

        def reset(self, token):
            # type: (_Token[T]) -> None
            if token._used:
                raise RuntimeError("_Token has already been used once")

            if token._var is not self:  # noqa
                raise ValueError("_Token was created by a different _ContextVar")

            if token._context is not _get_context():  # noqa
                raise ValueError("_Token was created in a different Context")

            ctx = token._context  # noqa
            if token._old_value is _Token.MISSING:  # noqa
                ctx._data = ctx._data.remove(token._var)  # noqa
            else:
                ctx._data = ctx._data.set(token._var, token._old_value)  # noqa

            token._used = True

        def __repr__(self):
            # type: () -> str
            r = "<ContextVar name={!r}".format(self.name)
            if self._default is not _NO_DEFAULT:
                r += " default={!r}".format(self._default)
            return r + " at {:0x}>".format(id(self))

    class _ContextMeta(GenericMeta):
        def __new__(mcs, names, bases, dct, **kwargs):
            cls = super(_ContextMeta, mcs).__new__(mcs, names, bases, dct, **kwargs)
            if cls.__module__ != _MODULE or cls.__name__ != "_Context":
                raise TypeError("type 'Context' is not an acceptable base type")
            return cls

    class _Context(six.with_metaclass(_ContextMeta, Mapping)):  # type: ignore
        __slots__ = ("_data", "_prev_context")

        def __init__(self):
            # type: () -> None
            self._data = pyrsistent.pmap()  # type: PMap[_ContextVar, Any]
            self._prev_context = None  # type: _Context | None

        def run(self, func, *args, **kwargs):
            # type: (Callable[..., RT], *Any, **Any) -> RT
            if self._prev_context is not None:
                raise RuntimeError("cannot enter context: {} is already entered".format(self))

            self._prev_context = _get_context()
            try:
                _set_context(self)
                return func(*args, **kwargs)
            finally:
                _set_context(self._prev_context)
                self._prev_context = None

        def copy(self):
            # type: () -> _Context
            new = _Context()
            new._data = self._data
            return new

        def __getitem__(self, var):
            # type: (_ContextVar[T]) -> T
            if not isinstance(var, _ContextVar):
                raise TypeError("a _ContextVar key was expected, got {!r}".format(var))
            return self._data[var]

        def __contains__(self, var):
            # type: (Any) -> bool
            if not isinstance(var, _ContextVar):
                raise TypeError("a _ContextVar key was expected, got {!r}".format(var))
            return var in self._data

        def __len__(self):
            # type: () -> int
            return len(self._data)

        def __iter__(self):
            # type: () -> Iterator[_ContextVar]
            for var in self._data:
                yield var

    class _TokenMeta(GenericMeta):
        def __new__(mcs, names, bases, dct, **kwargs):
            cls = super(_TokenMeta, mcs).__new__(mcs, names, bases, dct, **kwargs)
            if cls.__module__ != _MODULE or cls.__name__ != "_Token":
                raise TypeError("type '_Token' is not an acceptable base type")
            return cls

    class _Token(six.with_metaclass(_TokenMeta, Generic[T])):  # type: ignore
        __slots__ = ("__weakref__", "_context", "_var", "_old_value", "_used")

        MISSING = object()

        def __init__(self, context, var, old_value):
            # type: (_Context, _ContextVar[T], T) -> None
            self._context = context
            self._var = var
            self._old_value = old_value
            self._used = False

        @property
        def var(self):
            # type: () -> _ContextVar[T]
            return self._var

        @property
        def old_value(self):
            # type: () -> T
            return self._old_value

        def __repr__(self):
            # type: () -> str
            r = "<Token "
            if self._used:
                r += " used"
            r += " var={!r} at {:0x}>".format(self._var, id(self))
            return r

    def _copy_context():
        # type: () -> _Context
        return _get_context().copy()

    def _get_context():
        # type: () -> _Context
        ctx = getattr(_state, "context", None)
        if ctx is None:
            ctx = _Context()
            _state.context = ctx
        return ctx

    def _set_context(ctx):
        # type: (_Context) -> None
        _state.context = ctx

    _state = threading.local()

    type.__setattr__(_ContextVar, "__name__", "ContextVar")
    type.__setattr__(_ContextVar, "__qualname__", "ContextVar")
    globals()["ContextVar"] = _ContextVar

    type.__setattr__(_Context, "__name__", "Context")
    type.__setattr__(_Context, "__qualname__", "Context")
    globals()["Context"] = _Context

    type.__setattr__(_Token, "__name__", "Token")
    type.__setattr__(_Token, "__qualname__", "Token")
    globals()["Token"] = _Token

    object.__setattr__(_copy_context, "__name__", "copy_context")
    object.__setattr__(_copy_context, "__qualname__", "copy_context")
    globals()["copy_context"] = _copy_context



"""
Metaclass that forces `__hash__` to None when `__eq__` is declared.
This is a backport of the default behavior in
`Python 3 <https://docs.python.org/3/reference/datamodel.html#object.__hash__>`_.
"""

import six

__all__ = ["ImplicitHashMeta", "ImplicitHash"]


class ImplicitHashMeta(type):
    """Metaclass that forces `__hash__` to None when `__eq__` is declared."""

    @staticmethod
    def __new__(mcs, name, bases, dct, **kwargs):
        if "__eq__" in dct and "__hash__" not in dct:
            dct = dict(dct)
            dct["__hash__"] = None
        return super(ImplicitHashMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)


class ImplicitHash(six.with_metaclass(ImplicitHashMeta, object)):
    """Class that forces `__hash__` to None when `__eq__` is declared."""

    __slots__ = ()



import pickle

import pytest  # noqa
import six
from tippo import Type

from basicco.dynamic_class import make_cls
from basicco.obj_state import Reducible, ReducibleMeta


class MetaClass(ReducibleMeta):
    pass


class BaseClass(six.with_metaclass(ReducibleMeta, Reducible)):
    foo = None


class ParentClass(object):
    MyClass = None  # type: Type[BaseClass] | None


def test_make_cls():
    cls = make_cls(
        "ParentClass.MyClass",
        bases=(BaseClass,),
        meta=MetaClass,
        dct={"foo": "bar"},
    )  # type: Type[BaseClass]
    ParentClass.MyClass = cls

    assert cls is ParentClass.MyClass
    assert cls.__name__ == "MyClass"
    assert cls.__qualname__ == "ParentClass.MyClass"
    assert cls.__module__ == __name__
    assert cls.foo == "bar"
    assert issubclass(cls, BaseClass)
    assert isinstance(cls, MetaClass)
    assert type(pickle.loads(pickle.dumps(cls()))) is cls


if __name__ == "__main__":
    pytest.main()



import copy
import pickle

import pytest  # noqa

from basicco.hash_cache_wrapper import HashCacheWrapper


def test_pickling():
    cache = HashCacheWrapper(12345)
    assert cache == 12345

    assert pickle.loads(pickle.dumps(cache)) is None


def test_copy():
    cache = HashCacheWrapper(12345)
    assert cache == 12345

    assert copy.copy(cache) is None


def test_deepcopy():
    cache = HashCacheWrapper(12345)
    assert cache == 12345

    assert copy.deepcopy(cache) is None


if __name__ == "__main__":
    pytest.main()



"""Custom representation functions."""

from tippo import Any, Callable, Iterable, Mapping

__all__ = ["mapping_repr", "iterable_repr"]


def mapping_repr(
    mapping,  # type: Mapping | Iterable[tuple[Any, Any]]
    prefix="{",  # type: str
    template="{key}: {value}",  # type: str | Callable[..., str]
    separator=", ",  # type: str
    suffix="}",  # type: str
    sorting=False,  # type: bool
    sort_key=None,  # type: Callable[[tuple[Any, Any]], Any] | None
    reverse=False,  # type: bool
    key_repr=repr,  # type: Callable[[Any], str]
    value_repr=repr,  # type: Callable[[Any], str]
):
    # type: (...) -> str
    """
    Get custom representation of a mapping/items.

    :param mapping: Mapping.
    :param prefix: Prefix.
    :param template: Item format template or callable (supports {key}, {value}, {i}).
    :param separator: Separator.
    :param suffix: Suffix.
    :param sorting: Whether to sort keys.
    :param sort_key: Sorting key.
    :param reverse: Reverse sorting.
    :param key_repr: Key representation function.
    :param value_repr: Value representation function.
    :return: Custom representation.
    """
    if isinstance(mapping, Mapping):
        iterable = mapping.items()  # type: Iterable[tuple[Any, Any]]
    else:
        iterable = mapping

    if sort_key is None:
        sort_key = lambda item: item[0]
    if sorting:
        iterable = sorted(iterable, key=sort_key, reverse=reverse)

    if not callable(template):
        template = lambda _template=template, **v: _template.format(**v)

    parts = []  # type: list[str]
    for i, (key, value) in enumerate(iterable):
        part = template(key=key_repr(key), value=value_repr(value), i=i)
        parts.append(part)

    return prefix + separator.join(parts) + suffix


def iterable_repr(
    iterable,  # type: Iterable
    prefix="[",  # type: str
    template="{value}",  # type: str | Callable[..., str]
    separator=", ",  # type: str
    suffix="]",  # type: str
    sorting=False,  # type: bool
    sort_key=None,  # type: Callable[[Any], Any] | None
    reverse=False,  # type: bool
    value_repr=repr,  # type: Callable[[Any], str]
):
    # type: (...) -> str
    """
    Get custom representation of an iterable.

    :param iterable: Iterable.
    :param prefix: Prefix.
    :param template: Item format template or callable (supports {value}, {i}).
    :param separator: Separator.
    :param suffix: Suffix.
    :param sorting: Whether to sort the iterable or not.
    :param sort_key: Sorting key.
    :param reverse: Reverse sorting.
    :param value_repr: Value representation function.
    :return: Custom representation.
    """
    if sorting:
        iterable = sorted(iterable, key=sort_key, reverse=reverse)

    if not callable(template):
        template = lambda _template=template, **v: _template.format(**v)

    parts = []  # type: list[str]
    for i, value in enumerate(iterable):
        part = template(value=value_repr(value), i=i)
        parts.append(part)

    return prefix + separator.join(parts) + suffix



import pytest

from basicco.safe_not_equals import SafeNotEquals, SafeNotEqualsMeta


def test_class():
    assert isinstance(SafeNotEquals, SafeNotEqualsMeta)


def test_metaclass():
    obj_a = SafeNotEquals()
    obj_b = SafeNotEquals()
    assert (obj_a == obj_a) is not (obj_a != obj_a)
    assert (obj_b == obj_b) is not (obj_b != obj_b)
    assert (obj_a == obj_b) is not (obj_a != obj_b)


if __name__ == "__main__":
    pytest.main()



import collections

import pytest  # noqa
from tippo import NamedTuple

from basicco.named_tuple import defaults


def test_defaults():
    Point = collections.namedtuple("Point", ("x", "y"))
    defaults(x=0, y=0)(Point)
    assert Point() == (0, 0)  # noqa
    assert Point(1) == (1, 0)  # noqa
    assert Point(1, 2) == (1, 2)

    with pytest.raises(TypeError):
        defaults(x=0)(Point)(1)


def test_typed_defaults():
    Point = NamedTuple("Point", (("x", int), ("y", int)))  # noqa
    defaults(x=0, y=0)(Point)
    assert Point() == (0, 0)  # noqa
    assert Point(1) == (1, 0)  # noqa
    assert Point(1, 2) == (1, 2)  # noqa

    with pytest.raises(TypeError):
        defaults(x=0)(Point)(1)


if __name__ == "__main__":
    pytest.main()



"""An integer subclass that pickles/copies as None."""

__all__ = ["HashCacheWrapper"]


def _none_constructor():
    return None


class HashCacheWrapper(int):
    """An integer subclass that pickles/copies as None. This can be used to avoid serializing a cached hash value."""

    def __copy__(self):
        return None

    def __deepcopy__(self, memodict=None):
        return None

    def __reduce__(self):
        return _none_constructor, ()



"""Dynamic class generation."""

import types

import six
from tippo import Any, Iterable, Mapping, Type, TypeVar

from .caller_module import caller_module
from .init_subclass import InitSubclassMeta

__all__ = ["make_cls"]


T = TypeVar("T")


def make_cls(
    qualified_name,  # type: str
    bases=None,  # type: Iterable[type] | None
    meta=None,  # type: Type | None
    dct=None,  # type: Mapping[str, Any] | None
    kwargs=None,  # type: Mapping[str, Any] | None
    module=None,  # type: str | None
):
    # type: (...) -> Type[T]
    """
    Dynamically make a class.

    :param qualified_name: Qualified name.
    :param bases: Bases.
    :param meta: Metaclass.
    :param dct: Class body dictionary.
    :param kwargs: Class keyword arguments.
    :param module: Class module.
    :return: Generated class.
    """
    if bases is None:
        bases = (object,)
    else:
        bases = tuple(bases)

    if dct is None:
        dct_ = {}
    else:
        dct_ = dict(dct)

    if kwargs is None:
        kwargs = {}
    else:
        kwargs = dict(kwargs)

    if module is None:
        module = caller_module()

    name = qualified_name.split(".")[-1]
    dct_["__qualname__"] = qualified_name
    dct_["__module__"] = module

    if hasattr(types, "new_class"):

        if meta is not None:
            kwargs["metaclass"] = meta

        def exec_body(namespace):
            """Construct class body."""
            for key, value in six.iteritems(dct_):
                namespace[key] = value
            return namespace

        cls = types.new_class(name, bases, kwargs, exec_body)
    else:

        if meta is not None:
            bases = (six.with_metaclass(meta, *bases),)

        if kwargs and any(isinstance(b, InitSubclassMeta) for b in bases):
            dct_["__kwargs__"] = kwargs
            kwargs = {}

        cls = type(name, bases, dct_, **kwargs)

    type.__setattr__(cls, "__qualname__", qualified_name)
    type.__setattr__(cls, "__module__", module)

    return cls  # noqa



import pytest  # noqa
import six
from tippo import Generic, GenericMeta, TypeVar

from basicco.runtime_final import _FINAL_CLASS_TAG  # noqa
from basicco.runtime_final import _FINAL_METHODS  # noqa
from basicco.runtime_final import RuntimeFinal, RuntimeFinalMeta, final


def test_class():
    assert isinstance(RuntimeFinal, RuntimeFinalMeta)


def test_final_class():
    @final
    class Class(six.with_metaclass(RuntimeFinalMeta, object)):
        pass

    # Class should be tagged as final.
    assert getattr(Class, _FINAL_CLASS_TAG) is True

    # Should error when trying to subclass from final class.
    with pytest.raises(TypeError):

        class SubClass(Class):  # type: ignore
            pass

        assert not SubClass


def test_final_method():
    def dummy_decorator(func):
        return func

    class PropertyLikeDescriptor(object):
        def __init__(self, fget):
            self.fget = fget

        def __get__(self, instance, owner):
            return 3

    decorators = [
        dummy_decorator,
        PropertyLikeDescriptor,
        classmethod,
        staticmethod,
        property,
    ]

    # Different kinds of decorated members should be recognized as final.
    for decorator in decorators:

        class Class(six.with_metaclass(RuntimeFinalMeta, object)):
            @decorator
            @final
            def method(self):
                pass

        assert "method" in getattr(Class, _FINAL_METHODS)

        Class.new_method = decorator(final(lambda _: None))
        assert "new_method" in getattr(Class, _FINAL_METHODS)

        del Class.new_method
        assert "new_method" not in getattr(Class, _FINAL_METHODS)

        with pytest.raises(TypeError):

            class SubClass(Class):
                @decorator
                def method(self):  # type: ignore
                    pass

            assert not SubClass


def test_descriptor():
    class Descriptor(object):
        def __init__(self, func):
            self.func = func

        def __get__(self, instance, owner):
            return 3

    class Class(six.with_metaclass(RuntimeFinalMeta, object)):
        @final
        @Descriptor
        def method(self):
            pass

    # Decorated descriptor object should be recognized as final.
    assert "method" in getattr(Class, _FINAL_METHODS)


def test_generic():
    T = TypeVar("T")  # type: ignore  # noqa

    class BaseMeta(RuntimeFinalMeta, GenericMeta):
        pass

    class Base(six.with_metaclass(BaseMeta, Generic[T])):
        @final
        def method(self):
            pass

    assert Base[int]
    assert Base[T]

    assert repr(Base[int])
    assert repr(Base[T])

    class SubBase(Base[T]):
        pass

    assert SubBase


if __name__ == "__main__":
    pytest.main()



import copy
import pickle
import sys

import pytest  # noqa
import six

from basicco.obj_state import Reducible, ReducibleMeta, get_state, reducer, update_state


class _Class(object):
    __slots__ = ("a", "b", "c", "d")

    if sys.version_info[0:2] < (3, 4):
        __reduce__ = reducer


class _ReducibleClass(six.with_metaclass(ReducibleMeta, object)):
    __slots__ = ("a", "b", "c", "d")


class _SlottedClass(six.with_metaclass(ReducibleMeta, object)):
    __slots__ = ("a", "b")


class _MixedClass(_SlottedClass):
    c = None
    d = None


def test_get_state():
    class Class(object):
        def __init__(self):
            self.a = 1
            self.b = 2

    class SubClass(Class):
        def __init__(self):
            super(SubClass, self).__init__()
            self.c = 3

    assert get_state(SubClass()) == {"a": 1, "b": 2, "c": 3}


def test_get_slotted_state():
    class Class(object):
        __slots__ = ("a", "b")

        def __init__(self):
            self.a = 1
            self.b = 2

    class SubClass(Class):
        __slots__ = ("c",)

        def __init__(self):
            super(SubClass, self).__init__()
            self.c = 3

    assert get_state(SubClass()) == {"a": 1, "b": 2, "c": 3}


def test_get_protected_slotted_state():
    class Class(object):
        __slots__ = ("__a", "__b")

        def __init__(self):
            self.__a = 1
            self.__b = 2

    class SubClass(Class):
        __slots__ = ("__a", "__b", "__c")

        def __init__(self):
            super(SubClass, self).__init__()
            self.__a = 3
            self.__b = 4
            self.__c = 5

    assert get_state(SubClass()) == {
        "_Class__a": 1,
        "_Class__b": 2,
        "_SubClass__a": 3,
        "_SubClass__b": 4,
        "_SubClass__c": 5,
    }


def test_get_class_state():
    class Class(object):
        a = 1
        b = 2

    assert get_state(Class) == dict(Class.__dict__)


def test_update_state():
    class Class(object):
        pass

    obj = Class()
    obj.d = 4
    update_state(obj, {"a": 1, "b": 2, "c": 3})

    assert obj.__dict__ == {"a": 1, "b": 2, "c": 3, "d": 4}


def test_update_slotted_state():
    class Class(object):
        __slots__ = ("a", "b", "c", "d")

    obj = Class()
    obj.d = 4
    update_state(obj, {"a": 1, "b": 2, "c": 3})

    assert obj.a == 1
    assert obj.b == 2
    assert obj.c == 3
    assert obj.d == 4


def test_mixed_class():
    obj = _MixedClass()
    update_state(obj, {"a": 1, "b": 2, "c": 3, "d": 4})

    assert obj.a == 1
    assert obj.b == 2
    assert obj.c == 3
    assert obj.d == 4

    pickled_obj = pickle.loads(pickle.dumps(obj))

    assert pickled_obj.a == 1
    assert pickled_obj.b == 2
    assert pickled_obj.c == 3
    assert pickled_obj.d == 4

    copied_obj = copy.copy(obj)

    assert copied_obj.a == 1
    assert copied_obj.b == 2
    assert copied_obj.c == 3
    assert copied_obj.d == 4


def test_reducer():
    obj = _Class()
    obj.a = 1
    obj.b = 2
    obj.c = 3
    obj.d = 4

    pickled_obj = pickle.loads(pickle.dumps(obj))

    assert pickled_obj.a == 1
    assert pickled_obj.b == 2
    assert pickled_obj.c == 3
    assert pickled_obj.d == 4


def test_reducible_class():
    assert isinstance(Reducible, ReducibleMeta)


def test_reducible_meta():
    obj = _ReducibleClass()
    obj.a = 1
    obj.b = 2
    obj.c = 3
    obj.d = 4

    pickled_obj = pickle.loads(pickle.dumps(obj))

    assert pickled_obj.a == 1
    assert pickled_obj.b == 2
    assert pickled_obj.c == 3
    assert pickled_obj.d == 4


if __name__ == "__main__":
    pytest.main()



"""
Backport of the functionality of `__init_subclass__` from
`PEP 487 <https://peps.python.org/pep-0487/>`_ to Python 2.7.
"""

import sys

import six

from .get_mro import get_mro

__all__ = ["InitSubclassMeta", "InitSubclass"]


class InitSubclassMeta(type):
    """Metaclass that backports the functionality of `__init_subclass__` from PEP 487 to Python 2.7."""

    @staticmethod
    def __new__(mcs, name, bases, dct, **kwargs):

        # Keep a copy of the original kwargs -- they are sometimes used by generics.
        original_kwargs = dict(kwargs)

        # Wipe out kwargs if using older python.
        if sys.version_info[:3] < (3, 6):
            kwargs = dict()

        # Get compatibility kwargs (defined in the body of the class).
        dct_had_kwargs = "__kwargs__" in dct
        if dct_had_kwargs:
            compat_kwargs = dct["__kwargs__"]
        else:
            compat_kwargs = {}

        # Error out if the same kwarg is defined both in the body and in the class arguments.
        conflicting_kwargs = set(compat_kwargs).intersection(kwargs)
        if conflicting_kwargs:
            error = "conflicting class keyword arguments {}".format(
                ", ".join(sorted(repr(k) for k in conflicting_kwargs))
            )
            raise TypeError(error)

        # Merge kwargs.
        kwargs.update(compat_kwargs)

        # For older python versions.
        if sys.version_info[:3] < (3, 6):

            # Ensure classmethod.
            if "__init_subclass__" in dct and not isinstance(dct["__init_subclass__"], classmethod):
                dct = dict(dct)
                dct["__init_subclass__"] = classmethod(dct["__init_subclass__"])

            # Build class and pass original kwargs (for generics).
            cls = super(InitSubclassMeta, mcs).__new__(mcs, name, bases, dct, **original_kwargs)

            # Find '__init_subclass__' method.
            method = None
            method_owner = None
            for base in get_mro(cls):
                if base is cls:
                    continue
                method = base.__dict__.get("__init_subclass__")
                if method is not None:
                    method_owner = base
                    break

            # Found a method.
            if method is not None:
                assert method_owner is not None

                # Invalid base.
                if not isinstance(method_owner, InitSubclassMeta):
                    error = "base {!r} defines '__init_subclass__' but does not utilize {!r} as a metaclass".format(
                        method_owner.__name__, InitSubclassMeta.__name__
                    )
                    raise TypeError(error)

                # Call it with compat kwargs only.
                method.__func__(cls, **compat_kwargs)

            # Have kwargs but no method was found.
            elif kwargs:
                error = "object.__init_subclass__() takes no keyword arguments"
                raise TypeError(error)

        # We don't need to do anything for newer python versions, just pass the merged kwargs.
        else:
            cls = super(InitSubclassMeta, mcs).__new__(mcs, name, bases, dct, **kwargs)

        # Remove kwargs from the body of the class.
        if not dct_had_kwargs and hasattr(cls, "__kwargs__"):
            error = "one or more bases for class {!r} define {!r} but do not utilize {!r} as a metaclass".format(
                name, "__kwargs__", InitSubclassMeta.__name__
            )
            raise TypeError(error)
        elif dct_had_kwargs:
            type.__delattr__(cls, "__kwargs__")

        return cls


class InitSubclass(six.with_metaclass(InitSubclassMeta, object)):
    """Class that backports the functionality of `__init_subclass__` from PEP 487 to Python 2.7"""

    __slots__ = ()

    if sys.version_info[:3] < (3, 6):

        def __init_subclass__(cls):
            pass



from ._bases import (
    Base,
    BaseMeta,
    CompatBase,
    CompatBaseMeta,
    SlottedBase,
    SlottedBaseMeta,
    abstract,
    final,
    is_abstract,
    is_locked,
    set_locked,
    unlocked_context,
)

__all__ = [
    "CompatBaseMeta",
    "CompatBase",
    "BaseMeta",
    "Base",
    "SlottedBase",
    "SlottedBaseMeta",
    "abstract",
    "is_abstract",
    "final",
    "unlocked_context",
    "set_locked",
    "is_locked",
]


# Set module to this one.
for cls in (CompatBaseMeta, CompatBase, BaseMeta, Base, SlottedBaseMeta, SlottedBase):
    type.__setattr__(cls, "__module__", __name__)



"""Decorator that prevents `__repr__` methods from raising exceptions and return a default representation instead."""

import functools
import traceback

from tippo import Callable, overload

__all__ = ["default_alternative_repr", "safe_repr"]


_in_repr = False


def default_alternative_repr(obj):
    # type: (object) -> str
    message = [ln.strip(" ") for ln in traceback.format_exc().split("\n") if ln.strip(" ")][-1]
    return "{}; repr failed due to {!r}>".format(object.__repr__(obj)[:-1], message)


@overload
def safe_repr(
    maybe_func,  # type: Callable[[object], str]
    alternative_repr=default_alternative_repr,  # type: Callable[[object], str]
):
    # type: (...) -> Callable[[object], str]
    pass


@overload
def safe_repr(
    maybe_func,  # type: None
    alternative_repr=default_alternative_repr,  # type: Callable[[object], str]
):
    # type: (...) -> Callable[[Callable[[object], str]], Callable[[object], str]]
    pass


def safe_repr(maybe_func=None, alternative_repr=default_alternative_repr):
    """
    Decorate a representation method/function to prevent infinite recursion.

    :param maybe_func: The '__repr__' and/or '__str__' method/function or None.
    :param alternative_repr: Alternative repr function in case the decorated one fails.
    :return: Decorated method function or decorator.
    """

    def decorator(func):
        # type: (Callable[[object], str]) -> Callable[[object], str]

        @functools.wraps(func)
        def decorated(self, *args, **kwargs):
            global _in_repr
            before = _in_repr
            if not before:
                _in_repr = True
            try:
                return func(self, *args, **kwargs)  # noqa
            except Exception:  # noqa
                if before:
                    raise
                return alternative_repr(self, *args, **kwargs)  # noqa
            finally:
                if not before:
                    _in_repr = False

        return decorated

    # Return decorated or decorated.
    if maybe_func is not None:
        return decorator(maybe_func)
    else:
        return decorator



import pytest  # noqa
import six
from tippo import Generic, TypeVar

from basicco.get_mro import get_mro, preview_mro, resolve_origin


def test_resolve_origin():
    T = TypeVar("T")  # type: ignore  # noqa

    class Base(Generic[T]):
        pass

    assert resolve_origin(Base) is Base
    assert resolve_origin(Base[int]) is Base
    assert resolve_origin(Base[T]) is Base


def test_get_mro():
    T = TypeVar("T")  # type: ignore  # noqa

    class Base(Generic[T]):
        pass

    class SubBase(Base[T]):
        pass

    class SubSubBase(SubBase[T], Base[T]):
        pass

    assert get_mro(SubSubBase) == (SubSubBase, SubBase, Base, Generic, object)
    assert get_mro(SubSubBase[T]) == (SubSubBase, SubBase, Base, Generic, object)
    assert get_mro(SubSubBase[int]) == (SubSubBase, SubBase, Base, Generic, object)


def test_preview_mro():
    class M(type):
        pass

    class A(six.with_metaclass(M, object)):
        pass

    class B(A):
        pass

    class C(A):
        pass

    class D(B, C):
        pass

    assert preview_mro(B, C) == get_mro(D)[1:]


if __name__ == "__main__":
    pytest.main()



"""Retrieve the caller's module name."""

import functools
import inspect

import six
from tippo import Callable, Iterable, TypeVar

__all__ = ["caller_module", "auto_caller_module"]


T = TypeVar("T")


def caller_module(frames=0):
    # type: (int) -> str | None
    """
    Get caller module name if possible.

    :param frames: How many frames to go back.
    :return: Module name or None.
    """
    try:
        frame = inspect.stack()[2 + frames]
        module = inspect.getmodule(frame[0])
    except IndexError:
        return None
    else:
        if module is None:
            try:
                return frame[0].f_globals["__name__"]
            except (IndexError, KeyError):
                return None
        else:
            return module.__name__


def auto_caller_module(
    iterable_params=None,  # type: Iterable[str] | str | None
    single_params=None,  # type: Iterable[str] | str | None
):
    # type: (...) -> Callable[[T], T]
    """
    Make a decorator for a function that takes iterable/single keyword arguments that contain module paths.
    If no paths are provided, set the caller module as a path.

    :param iterable_params: Keyword arguments that are supposed to contain module paths.
    :param single_params: Keyword argument that are supposed to contain a single module path or None.
    :return: Decorator.
    """
    if isinstance(iterable_params, six.string_types):
        iterable_params_ = (iterable_params,)  # type: tuple[str, ...]
    elif iterable_params is None:
        iterable_params_ = ()
    else:
        iterable_params_ = tuple(iterable_params)

    if isinstance(single_params, six.string_types):
        single_params_ = (single_params,)  # type: tuple[str, ...]
    elif single_params is None:
        single_params_ = ()
    else:
        single_params_ = tuple(single_params)

    def decorator(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            _iterable_params = []
            _single_params = []
            for iterable_param in iterable_params_:
                values = tuple(kwargs.get(iterable_param, ()))
                if not values:
                    _iterable_params.append(iterable_param)
            for single_param in single_params_:
                value = kwargs.get(single_param, None)
                if value is None:
                    _single_params.append(single_param)

            if _iterable_params or _single_params:
                module = caller_module()
                for iterable_param in _iterable_params:
                    kwargs[iterable_param] = (module,)
                for single_param in _single_params:
                    kwargs[single_param] = module

            return func(*args, **kwargs)

        return decorated

    return decorator



import sys

import pytest  # noqa
import six
from tippo import Generic, GenericMeta, TypeVar

from basicco.init_subclass import InitSubclass, InitSubclassMeta

T_co = TypeVar("T_co", covariant=True)


def test_class():
    assert isinstance(InitSubclass, InitSubclassMeta)


def test_init_subclass():
    class A(InitSubclass):
        __kwargs__ = {}

        def __init_subclass__(cls, foo=None, bar=None, **kwargs):
            assert cls is not A
            assert foo == "{}1".format(cls.__name__)
            assert bar == "{}2".format(cls.__name__)
            super(A, cls).__init_subclass__(**kwargs)  # noqa

    assert A
    assert not hasattr(A, "__kwargs__")

    class B(A):
        __kwargs__ = {"foo": "B1", "bar": "B2"}

    assert B
    assert not hasattr(B, "__kwargs__")

    class C(B):
        __kwargs__ = {"foo": "C1", "bar": "C2"}

    assert C
    assert not hasattr(C, "__kwargs__")


def test_bad_base_kwargs():
    class A(object):
        __kwargs__ = {}

    with pytest.raises(TypeError):

        class B(InitSubclass, A):
            pass

        assert not B


def test_bad_base_init_subclass():
    if sys.version_info[:3] < (3, 6):

        class A(object):
            def __init_subclass__(cls, **kwargs):
                pass

        with pytest.raises(TypeError):

            class B(A, InitSubclass):
                pass

            assert not B
    else:

        class A(object):
            def __init_subclass__(cls, **kwargs):
                pass

        class B(A, InitSubclass):
            pass

        assert B


def test_newer_python():
    if sys.version_info[:3] >= (3, 6):
        code = """
class A(InitSubclass):

    def __init_subclass__(cls, foo=None, bar=None, **kwargs):
        assert cls is not A
        assert foo == "{}1".format(cls.__name__)
        assert bar == "{}2".format(cls.__name__)
        super(A, cls).__init_subclass__(**kwargs)

assert A
assert not hasattr(A, "__kwargs__")

class B(A, foo="B1"):
    __kwargs__ = {"bar": "B2"}

assert B
assert not hasattr(B, "__kwargs__")

class C(B, bar="C2"):
    __kwargs__ = {"foo": "C1"}

assert C
assert not hasattr(C, "__kwargs__")
"""
        exec(code, globals(), globals())


def test_generics():
    class MyMeta(InitSubclassMeta, GenericMeta):
        pass

    class Base(six.with_metaclass(MyMeta, InitSubclass)):
        pass

    class MyClass(Generic[T_co], Base):
        pass

    assert MyClass[int]

    class MySubClass(MyClass[T_co]):
        pass

    assert MySubClass[int]


if __name__ == "__main__":
    pytest.main()
